# %% tags=["remove-cell"]
ds1 = ds_model.copy()
ds1.cdo_command("setname,model")
ds1.run()
ds2 = ds_obs.copy()
ds2.cdo_command("setname,observation")
ds2.run()
ds_cor = nc.open_data([ds1.current[0], ds2.current[0]])
ds_cor.merge(match=["month"])
ds_cor.run()
ds_ts = ds_cor.copy()
ds_cor.cor_time("model", "observation")
title = f"Seasonal temporal correlation between {variable} for model and observations"
ds_cor.run()

# output to nc

out = f"../../oceanval_results/temporals/{variable}_cor_{source}.nc"
if not os.path.exists(os.path.dirname(out)):
    os.makedirs(os.path.dirname(out))
ds_cor.to_nc(out, zip = True, overwrite = True)
out_def = f"../../oceanval_results/temporals/{variable}_cor_{source}.pkl"
with open(out_def, "wb") as f:
    pickle.dump(definitions, f)


# %% tags=["remove-input"]

df_cor = ds_cor.to_dataframe().reset_index()
if fix_grid:
    ds_cor_plot = ds_cor.copy()
    ds_cor_plot.to_latlon(lon = [lon_min , lon_max], lat = [lat_min, lat_max], res = [lon_res, lat_res])
    ds_cor_plot.pub_plot("cor")
else:
    ds_cor.pub_plot("cor")



# %% tags=["remove-input"]
md(f"**Figure {i_figure}**: Seasonal temporal correlation between model and observations for {layer} {vv_name}. This is the Pearson correlation coefficient between climatological monthly mean values in the model and observations.")
i_figure += 1

# %% tags=["remove-cell"]
lon_range = lon_max - lon_min
lat_range = lat_max - lat_min

global_grid = False
if lon_range > 340:
    if lat_range > 160:
        global_grid = True


data_path = str(importlib.resources.files("oceanval").joinpath(f"data/{sub_regions}_subdomains.nc"))

regional = False

if sub_regions in ["nwes", "global"]: 
    ds_regions = nc.open_data(data_path, checks = False)
    if os.path.exists(data_path):
        regional = True
    # pull this in from the package data
    
    ds_regions.as_missing(0)
    ds_regions.set_fill(-9999)
    ds_regions.run()
    ds_regions.regrid(ds_model, method = "nn")
    regions_contents = ds_regions.contents
    

# %% tags=["remove-cell"]





# %% tags=["remove-cell"]
#shape = gpd.read_file(f"{data_dir}/mapping/TM_WORLD_BORDERS-0.3.shp")
mod_var = ds_model.variables[0]
obs_var = ds_obs.variables[0]
# create xlim using mod_var
try:
    time_name = [x for x in list(ds_model.to_xarray().coords) if "time" in x][0]
except:
    time_name = None
    concise = True
if len(ds_model.times) == 1:
    concise = True
    time_name = None
    model_unit = ""
    raw_extent = "foo"
    xlim = "bar"
    ylim = "ba"
if time_name is not None:
    lon_name = [x for x in list(ds_obs.to_xarray().coords) if "lon" in x][0]
    lat_name = [x for x in list(ds_obs.to_xarray().coords) if "lat" in x][0]

    df_model = (
        ds_model
        .to_dataframe()
        .reset_index()
        .rename(columns = {time_name: "time"})
        .rename(columns = {lon_name: "lon"})
        .rename(columns = {lat_name: "lat"})
        .assign(month = lambda x: x.time.dt.month)
        .loc[:,["lon", "lat", "month", mod_var ]]
        .rename(columns = {mod_var: "model"})
        .dropna()
    )

    xlim = np.array([df_model.lon.min(), df_model.lon.max()])
    ylim = np.array([df_model.lat.min(), df_model.lat.max()])

    df_obs = (
        ds_obs
        .to_dataframe()
        .reset_index()
    )

    time_name = [x for x in list(ds_obs.to_xarray().coords) if "time" in x][0]
    lon_name = [x for x in list(ds_obs.to_xarray().coords) if "lon" in x][0]
    lat_name = [x for x in list(ds_obs.to_xarray().coords) if "lat" in x][0]


    df_obs = (
        df_obs
        .rename(columns = {time_name: "time"}  )
        .rename(columns = {lon_name: "lon"}  )
        .rename(columns = {lat_name: "lat"}  )
        .assign(month = lambda x: x.time.dt.month)
        .loc[:,["lon", "lat", "month", obs_var ]]
        .rename(columns = {obs_var: "observation"})
        .dropna()

    )

    df_diff = (
        df_model
        .merge(df_obs, on = ["lon", "lat", "month"])
        .assign(diff = lambda x: x.model - x.observation)
    )
    try:
        model_unit = ds_model.contents.unit[0]
        model_unit = fix_unit(model_unit)
    except:
        model_unit = ds_obs.contents.unit[0]
        model_unit = fix_unit(model_unit)
    from oceanval.utils import get_extent
    raw_extent = get_extent(ds_annual[0])
    if np.abs(raw_extent[0] - df_model.lon.min()) > 3:
        # convert longitude to -180-180
        df_model["lon" ] = [x if x < 180 else x -360 for x in df_model.lon]
        df_obs["lon" ] = [x if x < 180 else x -360 for x in df_obs.lon]
        df_diff["lon" ] = [x if x < 180 else x -360 for x in df_diff.lon]

    # generate a temporary csv file name in /tmp
    # create adhoc dir if not
    if not os.path.exists("adhoc/tmp"):
        os.makedirs("adhoc/tmp")
    tmp_csv = f"adhoc/df_obs_model.csv"
    df_obs.to_csv(tmp_csv)
    tmp_csv = f"adhoc/df_model_model.csv"
    df_model.to_csv(tmp_csv)
    tmp_csv = f"adhoc/df_diff_model.csv"
    df_diff.to_csv(tmp_csv)


    # %% tags=["remove-input"]

    if not concise:
        md(f"The seasonal cycles of simulated and observed {vv_name} are compared in Figure {i_figure} below. This figure shows the model and observation average in each month of the year, and the differences between the two each month") 



# %% tags=["remove-input"]
# %%capture --no-display
# %%R -i model_unit  -w 800 -h 600 -i variable -i raw_extent -i concise -r 100


if(concise == FALSE){

fixed_scale = "fixed_scale_value"

if(fixed_scale == "False"){
    fixed_scale = FALSE
}
if(fixed_scale == "True"){
    fixed_scale = TRUE
}


library(tidyverse, warn.conflicts = FALSE)
library(cowplot, warn.conflicts = FALSE)
#library(tidyr, warn.conflicts = FALSE)
library(dplyr, warn.conflicts = FALSE)

df_model <- read_csv("adhoc/df_model_model.csv")


min_month = min(df_model$month)
if("month" %in% colnames(df_model) & min_month < 7){

    model_unit = str_replace(model_unit, "/m\\^3", "m<sup>-3</sup>")
    model_unit <- str_replace(model_unit, "/kg", "kg<sup>-1</sup>")

    df_model_raw <- drop_na(df_model)
    df_obs <- read_csv("adhoc/df_obs_model.csv")
    df_diff <- read_csv("adhoc/df_diff_model.csv")


#df_model_raw <- drop_na(df_model)
df_obs_raw <- drop_na(df_obs)
df_diff_raw <- drop_na(df_diff)

if (variable == "temperature" )
    model_unit = "°C"

df_model <- drop_na(df_model)
df_model <- df_model %>%
    filter(month %in% c(1,2,3,4,5,6))
df_obs <- df_obs %>%
    filter(month %in% c(1,2,3,4,5,6))
df_diff <- df_diff %>%
    filter(month %in% c(1,2,3,4,5,6))
# 

df_model_raw %>%
    group_by(month) %>%
    summarize(model_98 = quantile(model, probs = 0.98)) %>%
    ungroup() %>%
    summarize(model_98 = mean(model_98)) %>%
    ungroup() %>%
    pull(model_98) -> model_98

obs_98 <- df_obs_raw %>%
    group_by(month) %>%
    summarize(obs_98 = quantile(observation, probs = 0.98)) %>%
    ungroup() %>%
    summarize(obs_98 = mean(obs_98)) %>%
    ungroup() %>%
    pull(obs_98)

if(fixed_scale){
    model_98 = max(c(model_98, obs_98))
    obs_98 = max(c(model_98, obs_98))
    # calculate the 2nd percentile of model and obs
    model_02 <- df_model_raw %>%
        group_by(month) %>%
        summarize(model_02 = quantile(model, probs = 0.02)) %>%
        ungroup() %>%
        summarize(model_02 = mean(model_02)) %>%
        ungroup() %>%
        pull(model_02)
    obs_02 <- df_obs_raw %>%
        group_by(month) %>%
        summarize(obs_02 = quantile(observation, probs = 0.02)) %>%
        ungroup() %>%
        summarize(obs_02 = mean(obs_02)) %>%
        ungroup() %>%
        pull(obs_02)
    model_02 = min(c(model_02, obs_02))
    obs_02 = min(c(model_02, obs_02))

}

df_model <- df_model %>%
    mutate(model = ifelse(model > model_98, model_98, model))

df_obs <- df_obs %>%
    mutate(observation = ifelse(observation > obs_98, obs_98, observation))
if (fixed_scale){
    df_model <- df_model %>%
        mutate(model = ifelse(model < model_02, model_02, model))

    df_obs <- df_obs %>%
        mutate(observation = ifelse(observation < obs_02, obs_02, observation))
}

diff_02 <- df_diff_raw %>%
    group_by(month) %>%
    summarize(diff_02 = quantile(diff, probs = 0.02)) %>% 
    ungroup() %>%
    summarize(diff_02 = mean(diff_02)) %>%
    ungroup() %>%
    pull(diff_02)

diff_98 <- df_diff_raw %>%
    group_by(month) %>%
    summarize(diff_98 = quantile(diff, probs = 0.98)) %>%
    ungroup() %>%
    summarize(diff_98 = mean(diff_98)) %>%
    ungroup() %>%
    pull(diff_98)

df_diff <- df_diff %>% 
    mutate(diff = ifelse(diff < diff_02, diff_02, diff)) %>%
    mutate(diff = ifelse(diff > diff_98, diff_98, diff))



xlim = c(min(df_model$lon), max(df_model$lon))
ylim = c(min(df_model$lat), max(df_model$lat))

world_map <- map_data("world") 

# get 98th percentile of df_model$

# function to convert month number in int to month name
month_name <- function(x){
    month.abb[x]
}

# vectorize month_name function
# 
month_name <- Vectorize(month_name) 



# convert month number to month name in dataframes

# 

df_model <- df_model %>%
    mutate(month = month_name(month))

df_obs <- df_obs %>%
    mutate(month = month_name(month))

df_diff <- df_diff %>%
    mutate(month = month_name(month))

# first six months of the year in dataframes

df_model <- df_model %>%
    filter(month %in% c("Jan", "Feb", "Mar", "Apr", "May", "Jun"))

df_obs <- df_obs %>%
    filter(month %in% c("Jan", "Feb", "Mar", "Apr", "May", "Jun"))

df_diff <- df_diff %>%  
    filter(month %in% c("Jan", "Feb", "Mar", "Apr", "May", "Jun"))


# change month to suitable factor in dataframes

df_model$month <- factor(df_model$month, levels = c("Jan", "Feb", "Mar", "Apr", "May", "Jun"))

df_obs$month <- factor(df_obs$month, levels = c("Jan", "Feb", "Mar", "Apr", "May", "Jun"))

df_diff$month <- factor(df_diff$month, levels = c("Jan", "Feb", "Mar", "Apr", "May", "Jun"))


# Create sensible lon/lat labels based on the min/min lon and lat
# This needs to work on any data, including global data
lon_breaks = c(xlim[1], xlim[1] + 10, 0, xlim[2] - 10)
lat_breaks = c(ylim[1], ylim[1] + 5, ylim[1] + 10, ylim[1] + 15, ylim[2])
lon_labels = c(paste0(round(xlim[1]), "°W"), paste0(round(xlim[1] + 10), "°W"), "0°", paste0(round(xlim[2] - 10), "°E"))
lat_labels = c(paste0(round(ylim[1]), "°N"), paste0(round(ylim[1] + 5), "°N"), paste0(round(ylim[1] + 10), "°N"), paste0(round(ylim[1] + 15), "°N"), paste0(round(ylim[2]), "°N"))


gg1 = ggplot(df_model)+
    geom_tile(aes(x = lon, y = lat, fill = model))+
    facet_wrap(~month, ncol = 6)+
    coord_cartesian(xlim = xlim, ylim = ylim, expand = FALSE)+ 
    theme_bw(base_size = 12)+
    theme(legend.title = ggtext::element_markdown(angle = -90), legend.title.align = 0.5)+
    scale_fill_viridis_c(guide  = guide_colourbar(title.position = "right"))+
    labs(x = NULL, y = NULL, title = "Model", fill = model_unit)


gg1 <- gg1 +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 

gg1 <- gg1 + 
    geom_polygon(data = world_map, aes(x = long, y = lat, group = group), fill = "grey", colour = "grey")

gg2 = ggplot(df_obs)+
    geom_tile(aes(x = lon, y = lat, fill = observation))+
    facet_wrap(~month, ncol = 6)+
    coord_cartesian(xlim = xlim, ylim = ylim, expand = FALSE)+ 
    scale_fill_viridis_c(guide  = guide_colourbar(title.position = "right"))+
    theme_bw(base_size = 12)+
    theme(legend.title = ggtext::element_markdown(angle = -90), legend.title.align = 0.5)+
    labs(x = NULL, y = NULL, title = "Observation", fill = model_unit)


gg2 <- gg2 +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 

gg2 <-  gg2 + 
    geom_polygon(data = world_map, aes(x = long, y = lat, group = group), fill = "grey", colour = "grey")

gg3 = ggplot(df_diff)+
geom_tile(aes(x = lon, y = lat, fill = diff))+
facet_wrap(~month, ncol = 6)+
coord_cartesian(xlim = xlim, ylim = ylim, expand = FALSE)+ 
theme_bw(base_size = 12)+
scale_fill_gradient2(guide  = guide_colourbar(title.position = "right"), low = "blue", high = "red")+
theme(legend.title = ggtext::element_markdown(angle = -90), legend.title.align = 0.5)+
labs(x = NULL, y = NULL, title = "Model - Observation", fill = model_unit)


gg3 <- gg3 +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 


gg3 <- gg3 + 
        geom_polygon(data = world_map, aes(x = long, y = lat, group = group), fill = "grey", colour = "grey")


cowplot::plot_grid(gg1, gg2, gg3, ncol = 1)

}
}



# %% tags=["remove-input"]
# %%capture --no-display
# %%R -i model_unit -w 800 -h 600 -i variable -i raw_extent -i concise -r 100

if (concise == FALSE){

    fixed_scale = "fixed_scale_value"   
    if (fixed_scale == "False")
        fixed_scale = FALSE
    if (fixed_scale == "True")
        fixed_scale = TRUE
    
library(tidyverse, warn.conflicts = FALSE)
library(cowplot, warn.conflicts = FALSE)
library(dplyr, warn.conflicts = FALSE)

df_model <- read_csv("adhoc/df_model_model.csv")

max_month = max(df_model$month)
if("month" %in% colnames(df_model) & max_month > 6){

    df_model_raw <- drop_na(df_model)
    df_obs <- read_csv("adhoc/df_obs_model.csv")
    df_diff <- read_csv("adhoc/df_diff_model.csv")


df_obs_raw <- drop_na(df_obs)
df_diff_raw <- drop_na(df_diff)

if (variable == "temperature" )
    model_unit = "°C"

    model_unit = str_replace(model_unit, "/m\\^3", "m<sup>-3</sup>")
    # fix /kg to kg^-2
    model_unit <- str_replace(model_unit, "/kg", "kg<sup>-1</sup>")

df_model <- drop_na(df_model)
df_model <- df_model %>%
    filter(month > 6) 
df_obs <- df_obs %>%
    filter(month > 6) 
df_diff <- df_diff %>%
    filter(month > 6) 

df_model_raw %>%
    group_by(month) %>%
    summarize(model_98 = quantile(model, probs = 0.98)) %>%
    ungroup() %>%
    summarize(model_98 = mean(model_98)) %>%
    ungroup() %>%
    pull(model_98) -> model_98

obs_98 <- df_obs_raw %>%
    group_by(month) %>%
    summarize(obs_98 = quantile(observation, probs = 0.98)) %>%
    ungroup() %>%
    summarize(obs_98 = mean(obs_98)) %>%
    ungroup() %>%
    pull(obs_98)

if (fixed_scale){
    model_98 = max(c(model_98, obs_98))
    obs_98 = max(c(model_98, obs_98))
    # calculate the 2nd percentile of model and obs
    model_02 <- df_model_raw %>%
        group_by(month) %>%
        summarize(model_02 = quantile(model, probs = 0.02)) %>%
        ungroup() %>%
        summarize(model_02 = mean(model_02)) %>%
        ungroup() %>%
        pull(model_02)
    obs_02 <- df_obs_raw %>%
        group_by(month) %>%
        summarize(obs_02 = quantile(observation, probs = 0.02)) %>%
        ungroup() %>%
        summarize(obs_02 = mean(obs_02)) %>%
        ungroup() %>%
        pull(obs_02)

}


df_model <- df_model %>%
    mutate(model = ifelse(model > model_98, model_98, model))

    if (fixed_scale){
        df_model <- df_model %>%
            mutate(model = ifelse(model < model_02, model_02, model))
    }



df_obs <- df_obs %>%
    mutate(observation = ifelse(observation > obs_98, obs_98, observation))

if (fixed_scale){
    df_obs <- df_obs %>%
        mutate(observation = ifelse(observation < obs_02, obs_02, observation))
    }

diff_02 <- df_diff_raw %>%
    group_by(month) %>%
    summarize(diff_02 = quantile(diff, probs = 0.02)) %>% 
    ungroup() %>%
    summarize(diff_02 = mean(diff_02)) %>%
    ungroup() %>%
    pull(diff_02)

diff_98 <- df_diff_raw %>%
    group_by(month) %>%
    summarize(diff_98 = quantile(diff, probs = 0.98)) %>%
    ungroup() %>%
    summarize(diff_98 = mean(diff_98)) %>%
    ungroup() %>%
    pull(diff_98)

df_diff <- df_diff %>% 
    mutate(diff = ifelse(diff < diff_02, diff_02, diff)) %>%
    mutate(diff = ifelse(diff > diff_98, diff_98, diff))



xlim = c(min(df_model$lon), max(df_model$lon))
ylim = c(min(df_model$lat), max(df_model$lat))

world_map <- map_data("world") 

# get 98th percentile of df_model$

# function to convert month number in int to month name
month_name <- function(x){
    month.abb[x]
}

# vectorize month_name function
# 
month_name <- Vectorize(month_name) 



# convert month number to month name in dataframes

# 

df_model <- df_model %>%
    mutate(month = month_name(month))

df_obs <- df_obs %>%
    mutate(month = month_name(month))

df_diff <- df_diff %>%
    mutate(month = month_name(month))

# first six months of the year in dataframes


df_model <- df_model %>%    
    filter(month %in% c("Jul", "Aug", "Sep", "Oct", "Nov", "Dec"))

df_obs <- df_obs %>%
    filter(month %in% c("Jul", "Aug", "Sep", "Oct", "Nov", "Dec"))

df_diff <- df_diff %>%
    filter(month %in% c("Jul", "Aug", "Sep", "Oct", "Nov", "Dec"))

# change month to suitable factor in dataframes

df_model$month <- factor(df_model$month, levels = c("Jul", "Aug", "Sep", "Oct", "Nov", "Dec"))

df_obs$month <- factor(df_obs$month, levels = c("Jul", "Aug", "Sep", "Oct", "Nov", "Dec"))

df_diff$month <- factor(df_diff$month, levels = c("Jul", "Aug", "Sep", "Oct", "Nov", "Dec"))

# Create sensible lon/lat labels based on the min/min lon and lat
# This needs to work on any data, including global data
lon_breaks = c(xlim[1], xlim[1] + 10, 0, xlim[2] - 10)
lat_breaks = c(ylim[1], ylim[1] + 5, ylim[1] + 10, ylim[1] + 15, ylim[2])
lon_labels = c(paste0(round(xlim[1]), "°W"), paste0(round(xlim[1] + 10), "°W"), "0°", paste0(round(xlim[2] - 10), "°E"))
lat_labels = c(paste0(round(ylim[1]), "°N"), paste0(round(ylim[1] + 5), "°N"), paste0(round(ylim[1] + 10), "°N"), paste0(round(ylim[1] + 15), "°N"), paste0(round(ylim[2]), "°N"))



gg1 = ggplot(df_model)+
geom_tile(aes(x = lon, y = lat, fill = model))+
facet_wrap(~month, ncol = 6)+
coord_cartesian(xlim = xlim, ylim = ylim, expand = FALSE)+ 
theme_bw(base_size = 12)+
theme(legend.title = ggtext::element_markdown(angle = -90), legend.title.align = 0.5)+
scale_fill_viridis_c(guide  = guide_colourbar(title.position = "right"))+
labs(x = NULL, y = NULL, title = "Model", fill = model_unit)


gg1 <- gg1 + 
        geom_polygon(data = world_map, aes(x = long, y = lat, group = group), fill = "grey", colour = "grey")

gg1 <- gg1 +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 


gg2 = ggplot(df_obs)+
geom_tile(aes(x = lon, y = lat, fill = observation))+
facet_wrap(~month, ncol = 6)+
coord_cartesian(xlim = xlim, ylim = ylim, expand = FALSE)+ 
scale_fill_viridis_c(guide  = guide_colourbar(title.position = "right"))+
theme_bw(base_size = 12)+
theme(legend.title = ggtext::element_markdown(angle = -90), legend.title.align = 0.5)+
labs(x = NULL, y = NULL, title = "Observation", fill = model_unit)

gg2 <- gg2 + 
        geom_polygon(data = world_map, aes(x = long, y = lat, group = group), fill = "grey", colour = "grey")

gg2 <- gg2 +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 

gg3 = ggplot(df_diff)+
geom_tile(aes(x = lon, y = lat, fill = diff))+
facet_wrap(~month, ncol = 6)+
coord_cartesian(xlim = xlim, ylim = ylim, expand = FALSE)+ 
theme_bw(base_size = 12)+
scale_fill_gradient2(guide  = guide_colourbar(title.position = "right"), low = "blue", high = "red")+
theme(legend.title = ggtext::element_markdown(angle = -90), legend.title.align = 0.5)+
labs(x = NULL, y = NULL, title = "Model - Observation", fill = model_unit)


gg3 <- gg3 + 
        geom_polygon(data = world_map, aes(x = long, y = lat, group = group), fill = "grey", colour = "grey")

gg3 <- gg3 +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 

gg1 <- gg1 + 
   theme(plot.margin = margin(t = 0,  # Top margin
                             r = 0,  # Right margin
                             b = 0,  # Bottom margin
                             l = 0)) # Left ggmargin 

# figure out if it's a global file
if(abs(raw_extent[1] - raw_extent[2]) > 350){
    gg1 <- gg1 + 
    scale_x_continuous(breaks = c(-180, -90, 0, 90, 180), labels = c("180°W", "90°W", "0°", "90°E", "180°E"))+
    scale_y_continuous(breaks = c(-90, -45, 0, 45, 90), labels = c("90°S", "45°S", "0°", "45°N", "90°N"))

    gg2 <- gg2 +
    scale_x_continuous(breaks = c(-180, -90, 0, 90, 180), labels = c("180°W", "90°W", "0°", "90°E", "180°E"))+
    scale_y_continuous(breaks = c(-90, -45, 0, 45, 90), labels = c("90°S", "45°S", "0°", "45°N", "90°N"))

    gg3 <- gg3 +
    scale_x_continuous(breaks = c(-180, -90, 0, 90, 180), labels = c("180°W", "90°W", "0°", "90°E", "180°E"))+
    scale_y_continuous(breaks = c(-90, -45, 0, 45, 90), labels = c("90°S", "45°S", "0°", "45°N", "90°N"))

}

# appropriate plotting for northwest European Shelf
if((raw_extent[1] > -30) & (raw_extent[2] < 20)){
    gg1 <- gg1 + 
    scale_x_continuous(breaks = c(-20, -10, 0, 10), labels = c("20°W", "10°W", "0°", "10°E"))+
    scale_y_continuous(breaks = c(45, 50, 55, 60, 65), labels = c("45°N", "50°N", "55°N", "60°N", "65°N"))
    gg2 <- gg2 + 
    scale_x_continuous(breaks = c(-20, -10, 0, 10), labels = c("20°W", "10°W", "0°", "10°E"))+
    scale_y_continuous(breaks = c(45, 50, 55, 60, 65), labels = c("45°N", "50°N", "55°N", "60°N", "65°N"))
    gg3 <- gg3 +
    scale_x_continuous(breaks = c(-20, -10, 0, 10), labels = c("20°W", "10°W", "0°", "10°E"))+
    scale_y_continuous(breaks = c(45, 50, 55, 60, 65), labels = c("45°N", "50°N", "55°N", "60°N", "65°N"))
}

# gg1
# reduce the size of the plot
# options(repr.plot.width = 10, repr.plot.height = 3)
# gg1
cowplot::plot_grid(gg1, gg2, gg3, ncol = 1)

}
}


# %% tags=["remove-input"]
if not concise:
    fixed_scale = "fixed_scale_value"
    if fixed_scale == "False":
        md(f"**Figure {i_figure}**: Monthly mean {layer} {vv_name} for the model, observation and the difference between model and observations. For clarity, the maximum values are capped to the 98th percentiles") 
    else:
        md(f"**Figure {i_figure}**: Monthly mean {layer} {vv_name} for the model, observation and the difference between model and observations. For clarity, the maximum and minimum values are capped to cover the 98th and 2nd percentiles of both model and observations.")
    i_figure += 1



import pandas as pd
import shutil
import glob
import subprocess
import warnings
import nctoolkit as nc
nc.session_info["stamp"] = nc.session_info["stamp"] + "_ecoval_output_"
import copy
from oceanval.matchall import matchup
import dill
#from oceanval.fixers import tidy_name
from oceanval.session import session_info
import webbrowser
from oceanval.chunkers import add_chunks
import os
import re
from oceanval.fvcom import fvcom_preprocess
import importlib

from oceanval.parsers import Validator, definitions

def reset():
    # add docstring
    """
    Reset the matchup definitions to their default state.
    This function resets the matchup definitions used in oceanval to their default state.
    """
    definitions.reset()


notebook_dict = dict()


add_point_comparison = definitions.add_point_comparison
add_gridded_comparison = definitions.add_gridded_comparison
# add_gridded_comparison_2 = definitions.add_gridded_comparison_2

# loop through the keys and make sure all attributes are set
#for key in definitions.keys:
for key in session_info["keys"]:
    try:
        x = definitions[key].gridded
    except AttributeError:
        raise AttributeError(f"Variable '{key}' is missing 'gridded' attribute")
    try:
        x = definitions[key].point
    except AttributeError:
        raise AttributeError(f"Variable '{key}' is missing 'point' attribute")
    try:
        x = definitions[key].sources
    except AttributeError:
        raise AttributeError(f"Variable '{key}' is missing 'source' attribute")
    try:
        x = definitions[key].short_name
    except AttributeError:
        raise AttributeError(f"Variable '{key}' is missing 'short_name' attribute")
    try:
        x = definitions[key].long_name
    except AttributeError:
        raise AttributeError(f"Variable '{key}' is missing 'long_name' attribute")
# add a test to make sure definitions is complete


def fix_toc(concise = True, data_dir = None, out_dir = None):
    short_titles = dill.load(open(f"{data_dir}/oceanval_matchups/short_titles.pkl", "rb")) 
    paths = glob.glob(f"{out_dir}/oceanval_report/notebooks/*.ipynb")
    variables = dill.load(open(f"{data_dir}/oceanval_matchups/variables_matched.pkl", "rb"))
    variables.sort()

    vv_dict = dict()
    for vv in variables:
        vv_paths = [os.path.basename(x) for x in paths if f"_{vv}.ip" in x]
        if len(vv_paths) > 0:
            vv_dict[vv] = vv_paths
    # get summary docs
    ss_paths = [os.path.basename(x) for x in paths if "summary" in x]

    out = f"{out_dir}/oceanval_report/_toc.yml"

    # write line by line to out
    i_chapter = 1
    with open(out, "w") as f:
        # "format: jb-book"
        x = f.write("format: jb-book\n")
        x = f.write("root: intro\n")
        x = f.write("parts:\n")
        x = f.write(f"- caption: Introduction\n")
        x = f.write("  chapters:\n")
        x = f.write(f"  - file: notebooks/001_methods.ipynb\n")

        # open notebook and replace book_chapter with i_chapter

        # open notebook and replace book_chapter with i_chapter
        with open(f"{out_dir}/oceanval_report/notebooks/001_methods.ipynb", "r") as file:
            filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("book_chapter", str(i_chapter))

        with open(f"{out_dir}/oceanval_report/notebooks/001_methods.ipynb", "w") as file:
            file.write(filedata)
        i_chapter += 1

        x = f.write(f"- caption: Summaries\n")
        x = f.write("  chapters:\n")
        for ff in ss_paths:
            x = f.write(f"  - file: notebooks/{ff}\n")
            # open notebook and replace book_chapter with i_chapter
            with open(f"{out_dir}/oceanval_report/notebooks/{ff}", "r") as file:
                filedata = file.read()

            filedata = filedata.replace("book_chapter", str(i_chapter))

            # Replace the target string
            # Write the file out again
            with open(f"{out_dir}/oceanval_report/notebooks/{ff}", "w") as file:
                file.write(filedata)
            i_chapter += 1

        # loop over variables in each vv_dict
        # value is the file in the chapter section
        # key is the variable name, so is the section
        for vv in vv_dict.keys():
            # capitalize if not ph
            vv_out = short_titles[vv]

            x = f.write(f"- caption: {vv_out}\n")
            x = f.write("  chapters:\n")
            for ff in vv_dict[vv]:
                x = f.write(f"  - file: notebooks/{ff}\n")

                # open notebook and replace book_chapter with i_chapter
                with open(f"{out_dir}/oceanval_report/notebooks/{ff}", "r") as file:
                    filedata = file.read()

                # Replace the target string
                filedata = filedata.replace("book_chapter", str(i_chapter))

                # Write the file out again

                with open(f"{out_dir}/oceanval_report/notebooks/{ff}", "w") as file:
                    file.write(filedata)
                i_chapter += 1



def fix_toc_comparison():
    book_dir = "oceanval_comparison"

    out = f"{book_dir}/compare/_toc.yml"
    # write line by line to out
    with open(out, "w") as f:
        # "format: jb-book"
        x = f.write("format: jb-book\n")
        x = f.write("root: intro\n")
        x = f.write("parts:\n")
        x = f.write(f"- caption: Comparisons with gridded surface observations\n")
        x = f.write("  chapters:\n")
        x = f.write(f"  - file: notebooks/comparison_bias.ipynb\n")
        x = f.write(f"  - file: notebooks/comparison_spatial.ipynb\n")
        x = f.write(f"  - file: notebooks/comparison_seasonal.ipynb\n")
        x = f.write(f"  - file: notebooks/comparison_regional.ipynb\n")
        x = f.write(f"- caption: Comparisons with point observations\n")
        x = f.write("  chapters:\n")
        x = f.write(f"  - file: notebooks/comparison_point_surface.ipynb\n")




def validate(
    lon_lim=None,
    lat_lim=None,
    concise = True,
    variables="all",
    fixed_scale = False,
    region = None,
    data_dir = ".",
    out_dir = ".",
    test=False,
):
    # docstring
    """
    Run the model evaluation for all of the available datasets, and generate a validation report.

    Parameters
    ----------
    lon_lim : list or None
        The longitude limits for the validation. Default is None
    lat_lim : list or None
        The latitude limits for the validation. Default is None
    variables : str or list
        The variables to run the model evaluation for. Default is "all"
    fixed_scale : bool
        Whether to use a fixed scale for the seasonal plots. Default is False. If True, the minimum and maximum values are capped to cover the 2nd and 98th percentiles of both model and observations.
    region : str or None
        The region being validated. Must be either "nwes" (northwest European Shelf) or "global". Default is None.
    test : bool
        Default is False. Ignore, unless you are testing oceanval.

    Returns
    -------
    None
    """
    # convert data_dir to absolute path
    data_dir = os.path.expanduser(data_dir)
    data_dir = os.path.abspath(data_dir)
    #  regioncan only be nwes or global
    if region is not None:
        if region not in ["nwes", "global"]:
            raise ValueError("region must be either 'nwes' or 'global'")
    # ensure proper handling of ~
    out_dir = os.path.expanduser(out_dir)
    out_dir = os.path.abspath(out_dir)

    book_dir = os.path.join(out_dir, "oceanval_report")
    # if it doesn't exist, create it
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    # if lon_lim  is not None, make sure it's a list
    if lon_lim is not None:
        if isinstance(lon_lim, list) == False:
            raise ValueError("lon_lim must be a list")
        else:
            if len(lon_lim) != 2:
                raise ValueError("lon_lim must be a list of length 2")
    if lat_lim is not None:
        if isinstance(lat_lim, list) == False:
            raise ValueError("lat_lim must be a list")
        else:
            if len(lat_lim) != 2:
                raise ValueError("lat_lim must be a list of length 2")


    path_df = []

    fast_plot = False


    empty = True

    # book directory is book, book1, book2, book10 etc.

    # create a new name if one already exists
    i = 0

    if os.path.exists(book_dir):
        # get user input to decide if it should be removed
        while True:
                files = glob.glob(f"{book_dir}/**/**/**", recursive=True)
                # list all files in book, recursively
                for ff in files:
                    if ff.startswith(f"{book_dir}/"):
                        try:
                            os.remove(ff)
                        except:
                            pass
                files = glob.glob(f"{book_dir}/**/**/**", recursive=True)
                # only list files
                files = [x for x in files if os.path.isfile(x)]
                if len(files) == 0:
                    break

    # remove the results directory
    x_path = "oceanval_results"
    if os.path.exists(x_path):
        if x_path == "oceanval_results":
            shutil.rmtree(x_path)

    if variables != "all":
        if isinstance(variables, str):
            variables = [variables]

    if empty:
        from shutil import copyfile

        if not os.path.exists(book_dir):
            os.mkdir(book_dir)
        if not os.path.exists(f"{book_dir}/notebooks"):
            os.mkdir(f"{book_dir}/notebooks")

        data_path = importlib.resources.files(__name__).joinpath("data/001_methods.ipynb")
        if not os.path.exists(f"{book_dir}/notebooks/001_methods.ipynb"):
            copyfile(data_path, f"{book_dir}/notebooks/001_methods.ipynb")
        # open this file and replace model_name with model


        data_path = importlib.resources.files(__name__).joinpath("data/_toc.yml")

        out = f"{book_dir}/" + os.path.basename(data_path)
        copyfile(data_path, out)

        data_path = importlib.resources.files(__name__).joinpath("data/requirements.txt")
        out = f"{book_dir}/" + os.path.basename(data_path)
        copyfile(data_path, out)

        data_path = importlib.resources.files(__name__).joinpath("data/intro.md")
        out = f"{book_dir}/" + os.path.basename(data_path)
        copyfile(data_path, out)

        # copy config

        data_path = importlib.resources.files(__name__).joinpath("data/_config.yml")
        out = f"{book_dir}/" + os.path.basename(data_path)

        with open(data_path, "r") as file:
            filedata = file.read()


        # Write the file out again
        with open(out, "w") as file:
            file.write(filedata)

        # copyfile(data_path, out)

        path_df = []

        # loop through the point matchups and generate notebooks

        point_paths = glob.glob(f"{data_dir}/oceanval_matchups/point/**/**/**/**.csv")
        point_paths = [x for x in point_paths if "unit" not in os.path.basename(x)]
        # loop through the paths
        for pp in point_paths:
            ff_def = pp.replace(".csv", "_definitions.pkl")
            definitions = dill.load(open(ff_def, "rb"))
            vv = os.path.basename(pp).split("_")[2].replace(".csv", "")
            if variables != "all":
                if True:
                    if vv not in variables:
                        continue
            source = os.path.basename(pp).split("_")[0]
            variable = vv
            layer = os.path.basename(pp).split("_")[1].replace(".csv", "")
            Variable = definitions[variable].short_name

            vv_file = pp
            vv_file_find = pp.replace("../../", "")

            if os.path.exists(vv_file_find):
                if (
                    len(
                        glob.glob(
                            f"{book_dir}/notebooks/*point_{layer}_{variable}.ipynb"
                        )
                    )
                    == 0
                ):
                    file1 = importlib.resources.files(__name__).joinpath("data/point_template.ipynb")
                    with open(file1, "r") as file:
                        filedata = file.read()

                    if layer in [ "all", "surface"]:
                        filedata = filedata.replace(
                            "chunk_point_surface", "chunk_point"
                        )
                    else:
                        filedata = filedata.replace("chunk_point_surface", "")
                    if layer in ["bottom", "all"]:
                        if vv.lower() not in ["pco2"]:
                            filedata = filedata.replace(
                                "chunk_point_bottom", "chunk_point"
                            )
                        else:
                            filedata = filedata.replace("chunk_point_bottom", "")
                    else:
                        filedata = filedata.replace("chunk_point_bottom", "")

                    # Replace the target string
                    out = f"{book_dir}/notebooks/{source}_{layer}_{variable}.ipynb"
                    filedata = filedata.replace("point_variable", variable)
                    n_levels = definitions[variable].n_levels
                    if layer != "all":
                        if n_levels > 1:
                            filedata = filedata.replace(
                                "Validation of point_layer", f"Validation of {layer}"
                            )
                        else:
                            filedata = filedata.replace(
                                "Validation of point_layer", f"Validation of "
                            )
                    else:
                        filedata = filedata.replace(
                            "Validation of point_layer", f"Validation of "
                        )

                    filedata = filedata.replace("point_layer", layer)
                    filedata = filedata.replace("point_obs_source", source)
                    filedata = filedata.replace("template_title", Variable)
                    filedata = filedata.replace("data_dir_value", data_dir)
                    filedata = filedata.replace("out_dir_value", out_dir)

                    # Write the file out again
                    with open(out, "w") as file:
                        file.write(filedata)

                    path_df.append(
                        pd.DataFrame(
                            {
                                "variable": [variable],
                                "path": out,
                            }
                        )
                    )

        # Loop through the gridded matchups and generate notebooks
        # identify gridded variables in matched data
        gridded_paths = glob.glob(f"{data_dir}/oceanval_matchups/gridded/**/**.nc")

        if len(gridded_paths) > 0:
            for vv in [
                os.path.basename(x).split("_")[1].replace(".nc", "")
                for x in gridded_paths
            ]:
                for source in [
                    os.path.basename(x).split("_")[0]
                    for x in glob.glob(f"{data_dir}/oceanval_matchups/gridded/**/**_{vv}_**.nc")
                ]:

                    variable = vv
                    if variables != "all":
                        if vv not in variables:
                            continue
                    if not os.path.exists(
                        f"{book_dir}/notebooks/{source}_{variable}.ipynb"
                    ):
                        ff_def = glob.glob(f"{data_dir}/oceanval_matchups/gridded/{variable}/{source}_*definitions*.pkl")[0]
                        definitions = dill.load(open(ff_def, "rb"))
                        Variable = definitions[variable].short_name
                        ff_nc = glob.glob(f"{data_dir}/oceanval_matchups/gridded/{variable}/{source}_*surface*.nc")[0]
                        ds = nc.open_data(ff_nc, checks = False)
                        try:
                            n_months = len(ds.months)
                        except:
                            n_months = 12
                        seasonal = n_months >= 12
    
                        file1 = importlib.resources.files(__name__).joinpath("data/gridded_template.ipynb")
                        if (
                            len(
                                glob.glob(
                                    f"{book_dir}/notebooks/*{source}_{variable}.ipynb"
                                )
                            )
                            == 0
                        ):
                            with open(file1, "r") as file:
                                filedata = file.read()

                            # Replace the target string
                            filedata = filedata.replace("template_variable", variable)
                            filedata = filedata.replace("template_title", Variable)
                            filedata = filedata.replace("data_dir_value", data_dir)
                            filedata = filedata.replace("source_name", source)
                            if region == "nwes":
                                filedata = filedata.replace("zonal_height", "6000" )
                            else:
                                filedata = filedata.replace("zonal_height", "2000" )
                            # make every letter a capital
                            source_capital = source.upper()
                            filedata = filedata.replace("source_title", source_capital)
                            if seasonal is False:
                                filedata = filedata.replace(
                                    "chunk_seasonal", ""
                                )
                            # change sub_regions_value to region
                            if region is not None: 
                                filedata = filedata.replace(
                                    "sub_regions_value", str(region)
                                )

                            # Write the file out again
                            with open(
                                f"{book_dir}/notebooks/{source}_{variable}.ipynb", "w"
                            ) as file:
                                file.write(filedata)

                            variable = vv
                            path_df.append(
                                pd.DataFrame(
                                    {
                                        "variable": [variable],
                                        "path": [
                                            f"{book_dir}/notebooks/{source}_{variable}.ipynb"
                                        ],
                                    }
                                )
                            )

        # need to start by figuring out whether anything has already been run...

        i = 0

        for ff in [
            x for x in glob.glob("{book_dir}/notebooks/*.ipynb") if "info" not in x
        ]:
            try:
                i_ff = int(os.path.basename(ff).split("_")[0])
                if i_ff > i:
                    i = i_ff
            except:
                pass

        i_orig = i

        if len(path_df) > 0:
            path_df = pd.concat(path_df)
            path_df = path_df.sort_values("variable").reset_index(drop=True)

        for i in range(len(path_df)):
            file1 = path_df.path.values[i]
            # pad i with zeros using zfill
            i_pad = str(i + 1).zfill(3)
            new_file = (
                os.path.dirname(file1) + "/" + i_pad + "_" + os.path.basename(file1)
            )
            os.rename(file1, new_file)
            # print(key, value)

        # copy the summary.ipynb notebook and add i_pad to the name

        i = i + 2
        i_pad = str(i).zfill(3)



        file1 = importlib.resources.files(__name__).joinpath("data/summary.ipynb")
        if len(glob.glob(f"{book_dir}/notebooks/*summary.ipynb")) == 0:
            copyfile(file1, f"{book_dir}/notebooks/{i_pad}_summary.ipynb")
        else:
            if i > (i_orig + 1):
                initial = glob.glob(f"{book_dir}/notebooks/*summary.ipynb")[0]
                copyfile(initial, f"{book_dir}/notebooks/{i_pad}_summary.ipynb")
                i += 1

        # change domain_title to "Full domain"

        with open(f"{book_dir}/notebooks/{i_pad}_summary.ipynb", "r") as file:
            filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("domain_title", "Full domain")
        filedata = filedata.replace("data_dir_value", data_dir)
        filedata = filedata.replace("out_dir_value", out_dir)

        # Write the file out again
        with open(f"{book_dir}/notebooks/{i_pad}_summary.ipynb", "w") as file:
            file.write(filedata)

        # pair the notebooks using jupyter text

        os.system(
            f"jupytext --set-formats ipynb,py:percent {book_dir}/notebooks/*.ipynb"
        )

        # add the chunks
        add_chunks(out_dir)

        # loop through the notebooks and set r warnings options
        for ff in glob.glob(f"{book_dir}/notebooks/*.py"):
            with open(ff, "r") as file:
                filedata = file.read()

            # loop through line by line, and rewrite the original file
            lines = filedata.split("\n")
            new_lines = []
            for line in lines:
                if "%%R" in line:
                    new_lines.append(line)
                    new_lines.append("options(warn=-1)")
                else:
                    new_lines.append(line)
            # loop through all lines in lines and replace the_test_status with True
            for i in range(len(new_lines)):
                new_lines[i] = new_lines[i].replace("latexpagebreak", "")
                if "the_test_status" in new_lines[i]:
                    if test:
                        new_lines[i] = new_lines[i].replace("the_test_status", "True")
                    else:
                        new_lines[i] = new_lines[i].replace("the_test_status", "False")
                if '"gam"' in new_lines[i]:
                    new_lines[i] = new_lines[i].replace('"gam"', '"lm"')

                new_lines[i] = new_lines[i].replace("the_lon_lim", str(lon_lim))
                new_lines[i] = new_lines[i].replace("the_lat_lim", str(lat_lim))
                new_lines[i] = new_lines[i].replace("fixed_scale_value", str(fixed_scale))
                # replace concice_value with concice
                if "concise_value" in new_lines[i]:
                    if concise:
                        new_lines[i] = new_lines[i].replace(
                            "concise_value", "True"
                        )
                    else:
                        new_lines[i] = new_lines[i].replace("concise_value", "False")

            # write the new lines to the file
            with open(ff, "w") as file:
                for line in new_lines:
                    file.write(line + "\n")

        # sync the notebooks
        #
        os.system(f"jupytext --sync {book_dir}/notebooks/*.ipynb")


    # loop through notebooks and change fast_plot_value to fast_plot

    for ff in glob.glob(f"{book_dir}/notebooks/*.ipynb"):
        with open(ff, "r") as file:
            filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("fast_plot_value", str(fast_plot))
        filedata = filedata.replace("data_dir_value", data_dir)
        filedata = filedata.replace("out_dir_value", out_dir)

        # fix linees using the above
        def fix_r_magic(x):
            if "%%R" in x:
                x = re.sub(r" -r\s+\d+", "", x)
                x = x.replace("%%R", "%%R -r 120 ")
            return x

        # Write the file out again
        with open(ff, "w") as file:
            file.write(filedata)

    # fix the toc using the function

    fix_toc(concise = concise, data_dir = data_dir, out_dir = out_dir)

    for ff in glob.glob(f"{book_dir}/notebooks/*.ipynb"):
        ff_clean = ff.replace(".ipynb", ".py")
        if os.path.exists(ff_clean):
            os.remove(ff_clean)

    # move pml_logo to book directory

    shutil.copyfile(
        importlib.resources.files(__name__).joinpath("data/pml_logo.jpg"), f"{book_dir}/pml_logo.jpg"
    )

    os.system(f"jupyter-book build  {book_dir}/")

    stamps = [
        os.path.basename(x) for x in glob.glob(f"{book_dir}/notebooks/.trackers/*")
    ]
    stamps.append("nctoolkit_rwi_uhosarcenctoolkittmp")

    delete = []
    for x in stamps:
        delete += glob.glob("/tmp/*" + x + "*")

    for ff in delete:
        if os.path.exists(ff):
            if "nctoolkit" in x:
                os.remove(ff)

    out_ff = f"{book_dir}/_build/html/index.html"

    # create a symlink to the html file
    if os.path.exists(f"{out_dir}/oceanval_report.html"):
        os.remove(f"{out_dir}/oceanval_report.html")
    os.symlink(f"{book_dir}/_build/html/index.html", f"{out_dir}/oceanval_report.html")
    webbrowser.open(
        "file://" + os.path.abspath(f"{book_dir}/_build/html/index.html")
    )


def rebuild():
    """
    Rebuild the validation report after modifying notebooks.
    Use this if you have modified the notebooks generated and want to create a new validation report.

    Parameters
    ----------
    build : str
        The type of the existing build. Default is None. Options are "html" or "pdf"
    """
    # add a deprecation notice

    os.system(f"jupyter-book build oceanval_report/")

    webbrowser.open(
        "file://" + os.path.abspath(f"oceanval_report/_build/html/index.html")
    )




try:
    from importlib.metadata import version as _version
except ImportError:
    from importlib_metadata import version as _version

try:
    __version__ = _version("oceanval")
except Exception:
    __version__ = "999"


def compare(model_dict=None):
    """
    Compare pre-validated simulations.
    This function will compare the validation output from two simulations.

    Parameters
    ----------
    model_dict : dict
        A dictionary of model names and the paths to the validation output. Default is None.
        Example: {"model1": "/path/to/model1", "model2": "/path/to/model2"}
        If the models have different grids, put the model with the smallest grid first.

    """

    for key in model_dict.keys():
        # check that the path exists
        if not os.path.exists(model_dict[key]):
            raise ValueError(f"Path {model_dict[key]} does not exist")
        # convert to an absolute path
        model_dict[key] = os.path.abspath(model_dict[key])

    if os.path.exists("oceanval_comparison"):
        # get user input to decide if it should be removed
        user_input = input(
            "oceanval_comparison directory already exists. This will be emptied and replaced. Do you want to proceed? (y/n): "
        )
        if user_input.lower() == "y":
            while True:
                files = glob.glob("oceanval_comparison/**/**/**", recursive=True)
                # list all files in oceanval_comparison, recursively
                for ff in files:
                    if ff.startswith("oceanval_comparison"):
                        try:
                            os.remove(ff)
                        except:
                            pass
                files = glob.glob("oceanval_comparison/**/**/**", recursive=True)
                # only list files
                files = [x for x in files if os.path.isfile(x)]
                if len(files) == 0:
                    break
        else:
            print("Exiting")
            return None


    if not os.path.exists("oceanval_comparison/compare"):
        # create directory recursively
        os.makedirs("oceanval_comparison/compare")

    # copy the pml logo
    shutil.copyfile(
        importlib.resources.files(__name__).joinpath("data/pml_logo.jpg"),
        "oceanval_comparison/pml_logo.jpg",
    )

    # move toc etc to oceanval_comparison/compare

    data_path = importlib.resources.files(__name__).joinpath("data/_toc.yml")

    out = "oceanval_comparison/compare/" + os.path.basename(data_path)
    shutil.copyfile(data_path, out)

    fix_toc_comparison()

    data_path = importlib.resources.files(__name__).joinpath("data/requirements.txt")

    out = "oceanval_comparison/compare/" + os.path.basename(data_path)

    shutil.copyfile(data_path, out)

    data_path = importlib.resources.files(__name__).joinpath("data/intro.md")

    out = "oceanval_comparison/compare/" + os.path.basename(data_path)

    shutil.copyfile(data_path, out)

    # copy config

    data_path = importlib.resources.files(__name__).joinpath("data/_config.yml")

    out = "oceanval_comparison/compare/" + os.path.basename(data_path)

    shutil.copyfile(data_path, out)
    # read in out and change some values
    with open(out, "r") as file:
        filedata = file.read()
    # Replace the target string
    filedata = filedata.replace("oceanval_report", "../oceanval_comparison")
    # Write the file out again
    with open(out, "w") as file:
        file.write(filedata)


    # copy the comparison_seasonal notebook

    # make sure the directory exists

    if not os.path.exists("oceanval_comparison/compare/notebooks"):
        # create directory recursively
        os.makedirs("oceanval_comparison/compare/notebooks")

    file1 = importlib.resources.files(__name__).joinpath("data/comparison_seasonal.ipynb")
    if len(glob.glob("oceanval_comparison/compare/notebooks/*comparison_seasonal.ipynb")) == 0:
        shutil.copyfile(file1, "oceanval_comparison/compare/notebooks/comparison_seasonal.ipynb")


    model_dict_str = str(model_dict)

    with open("oceanval_comparison/compare/notebooks/comparison_seasonal.ipynb", "r") as file:
        filedata = file.read()

    # Replace the target string
    filedata = filedata.replace("model_dict_str", model_dict_str)

    # Write the file out again

    with open("oceanval_comparison/compare/notebooks/comparison_seasonal.ipynb", "w") as file:
        file.write(filedata)

    # now sort out the comparison_spatial notebook

    file1 = importlib.resources.files(__name__).joinpath("data/comparison_spatial.ipynb")
    if len(glob.glob("oceanval_comparison/compare/notebooks/*comparison_spatial.ipynb")) == 0:
        shutil.copyfile(file1, "oceanval_comparison/compare/notebooks/comparison_spatial.ipynb")

    model_dict_str = str(model_dict)

    with open("oceanval_comparison/compare/notebooks/comparison_spatial.ipynb", "r") as file:
        filedata = file.read()

    # Replace the target string
    filedata = filedata.replace("model_dict_str", model_dict_str)

    # Write the file out again

    with open("oceanval_comparison/compare/notebooks/comparison_spatial.ipynb", "w") as file:
        file.write(filedata)

    # move the regional book

    file1 = importlib.resources.files(__name__).joinpath("data/comparison_regional.ipynb")
    if len(glob.glob("oceanval_comparison/compare/notebooks/*comparison_regional.ipynb")) == 0:
        shutil.copyfile(file1, "oceanval_comparison/compare/notebooks/comparison_regional.ipynb")

    model_dict_str = str(model_dict)

    with open("oceanval_comparison/compare/notebooks/comparison_regional.ipynb", "r") as file:
        filedata = file.read()

    # Replace the target string
    filedata = filedata.replace("model_dict_str", model_dict_str)

    # Write the file out again

    with open("oceanval_comparison/compare/notebooks/comparison_regional.ipynb", "w") as file:
        file.write(filedata)

    # now to comparison_bias

    file1 = importlib.resources.files(__name__).joinpath("data/comparison_bias.ipynb")

    if len(glob.glob("oceanval_comparison/compare/notebooks/*comparison_bias.ipynb")) == 0:
        shutil.copyfile(file1, "oceanval_comparison/compare/notebooks/comparison_bias.ipynb")

    model_dict_str = str(model_dict)

    with open("oceanval_comparison/compare/notebooks/comparison_bias.ipynb", "r") as file:
        filedata = file.read()

    # Replace the target string
    filedata = filedata.replace("model_dict_str", model_dict_str)

    # Write the file out again

    with open("oceanval_comparison/compare/notebooks/comparison_bias.ipynb", "w") as file:
        file.write(filedata)

    # figure out if both simulations have point data

    i = 0

    if i == 0:
        for ss in [ "surface"]:
            file1 = importlib.resources.files(__name__).joinpath("data/comparison_point.ipynb")

            if (
                len(glob.glob(f"oceanval_comparison/compare/notebooks/*comparison_point_{ss}.ipynb"))
                == 0
            ):
                shutil.copyfile(
                    file1, f"oceanval_comparison/compare/notebooks/comparison_point_{ss}.ipynb"
                )

            model_dict_str = str(model_dict)

            with open(
                f"oceanval_comparison/compare/notebooks/comparison_point_{ss}.ipynb", "r"
            ) as file:
                filedata = file.read()

            # Replace the target string
            filedata = filedata.replace("model_dict_str", model_dict_str)

            # Write the file out again

            with open(
                f"oceanval_comparison/compare/notebooks/comparison_point_{ss}.ipynb", "w"
            ) as file:
                file.write(filedata)
            # replace layer in the notebook with ss
            with open(
                f"oceanval_comparison/compare/notebooks/comparison_point_{ss}.ipynb", "r"
            ) as file:
                filedata = file.read()

            # Replace the target string
            filedata = filedata.replace("layer", ss)

            # Write the file out again

            with open(
                f"oceanval_comparison/compare/notebooks/comparison_point_{ss}.ipynb", "w"
            ) as file:
                file.write(filedata)


    # sync the notebooks

    os.system("jupytext --set-formats ipynb,py:percent oceanval_comparison/compare/notebooks/*.ipynb")

    add_chunks()

    # replace the test status in the notebooks
    books = glob.glob("oceanval_comparison/compare/notebooks/*.py")
    for book in books:
        with open(book, "r") as file:
            filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("the_test_status", "False")

        # Write the file out again
        with open(book, "w") as file:
            file.write(filedata)

    # fix the chunks
    os.system("jupytext --sync oceanval_comparison/compare/notebooks/*.ipynb")

    # loop through notebooks and change fast_plot_value to fast_plot

    for ff in glob.glob("oceanval_comparison/compare/notebooks/*.ipynb"):
        with open(ff, "r") as file:
            filedata = file.read()

        # Replace the target string
        filedata = filedata.replace("fast_plot_value", "False")

        # Write the file out again
        with open(ff, "w") as file:
            file.write(filedata)

    os.system("jupyter-book build oceanval_comparison/compare/")
    import webbrowser


    webbrowser.open("file://" + os.path.abspath("oceanval_comparison/compare/_build/html/index.html"))

import platform
import tempfile


def temp_check():
    """
    Function to check temp files
    """

    if platform.system() == "Linux":
        mylist = [f for f in glob.glob("/tmp/*")]
        mylist = mylist + [f for f in glob.glob("/var/tmp/*")]
        mylist = mylist + [f for f in glob.glob("/usr/tmp/*")]
        mylist = [f for f in mylist if "nctoolkit" in f]

        mylist = [x for x in mylist if "ecoval_output" in x]

        session_info["old_files"] = mylist

        if len(mylist) > 0:
            if len(mylist) == 1:
                print(
                    f"{len(mylist)} temporary file was created by oceanval in prior or current "
                    f"sessions. Consider running oceanval.deep_clean!"
                )
            else:
                print(
                    f"{len(mylist)} temporary files were created by oceanval in prior or current"
                    f" sessions. Consider running oceanval.deep_clean!"
                )

    else:
        temp_folder = tempfile.gettempdir()

        mylist = [f for f in glob.glob(f"{temp_folder}/*")]
        mylist = [f for f in mylist if "nctoolkit" in f]
        mylist = [x for x in mylist if "ecovaloutput" in x]

        if len(mylist) > 0:
            if len(mylist) == 1:
                print(
                    f"{len(mylist)} temporary file was created by oceanval in prior or current "
                    f"sessions. Consider running oceanval.deep_clean!"
                )
            else:
                print(
                    f"{len(mylist)} temporary files were created by oceanval in prior or "
                    f"current sessions. Consider running oceanval.deep_clean!"
                )

temp_check()




def deep_clean():
    """
    Deep temp file cleaner.
    Remove all temporary files ever created by oceanval
    across all previous and current sesions
    """

    candidates = session_info["old_files"] 

    mylist = [f for f in candidates if "nctoolkit" in f and "ecoval_output" in f]
    for ff in mylist:
        if ff in session_info["old_files"]:
            if "nctoolkit" in ff and "ecoval_output" in ff:
                os.remove(ff)


import os
import nctoolkit as nc
import warnings
import xarray as xr
import numpy as np
import subprocess
import pandas as pd
from tqdm import tqdm
from oceanval.session import session_info


def bin_value(x, bin_res):
    return np.floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2


def extension_of_directory(starting_directory, exclude=[]):
    levels = session_info["levels_down"]

    new_directory = ""
    for i in range(levels):
        new_directory = new_directory + "/**"
    return new_directory + "/"


def is_latlon(ff):

    cdo_result = subprocess.run(
        f"cdo griddes {ff}",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    return "lonlat" in cdo_result.stdout.decode("utf-8")


def get_extent(ff):
    # add docstring
    """ "
    Get the extent of a netcdf file

    Parameters
    ----------
    ff : str
        The path to the netcdf file

    Returns
    -------
    extent : list
        A list of the form [lon_min, lon_max, lat_min, lat_max]

    """

    ds = nc.open_data(ff)
    ds.subset(variables=ds.variables[0])
    ds.top()
    ds.tmean()
    ds.as_missing(0)
    ds_xr = ds.to_xarray()
    lon_name = [x for x in ds_xr.coords if "lon" in x][0]
    lat_name = [x for x in ds_xr.coords if "lat" in x][0]

    df = ds_xr.to_dataframe().reset_index()
    df = ds.to_dataframe().dropna().reset_index()
    df = df.rename(columns={lon_name: "lon", lat_name: "lat"})
    df = df.dropna()
    max_lon = df["lon"].max()
    if max_lon > 180:
        new_lon = df["lon"].values % 360
        new_lon = new_lon - 180
        lon_min = float(new_lon.min())
        lon_max = float(new_lon.max())
    else:
        lon_min = float(df.lon.min())
        lon_max = float(df.lon.max())
    lat_min = float(df.lat.min())
    lat_max = float(df.lat.max())
    return [lon_min, lon_max, lat_min, lat_max]


def get_resolution(ff):
    ds = nc.open_data(ff, checks=False)
    var = ds.variables[0]
    ds = xr.open_dataset(ff)
    lon_name = [x for x in list(ds.coords) if "lon" in x][0]
    lat_name = [x for x in list(ds.coords) if "lat" in x][0]
    var_dims = ds[var].dims
    extent = get_extent(ff)
    if lon_name in var_dims:
        if lat_name in var_dims:
            n_lon = len(ds[lon_name])
            n_lat = len(ds[lat_name])
            lon_max = float(ds[lon_name].max())
            lon_min = float(ds[lon_name].min())
            lat_max = float(ds[lat_name].max())
            lat_min = float(ds[lat_name].min())
            lon_res = (lon_max - lon_min) / (n_lon - 1)
            lat_res = (lat_max - lat_min) / (n_lat - 1)
            return [lon_res, lat_res]
    else:
        # get the final two var_dims
        var_dims = var_dims[-2:]
        # lat should be the second
        n_lat = len(ds[var_dims[1]])
        n_lon = len(ds[var_dims[0]])
        lon_res = (extent[1] - extent[0]) / n_lon
        lat_res = (extent[3] - extent[2]) / n_lat
        return [lon_res, lat_res]


def fvcom_regrid(ff=None, new_grid=None, vv=None, lons=None, lats=None, res=None):
    # multiple = False
    # if len(vv) > 1:
    #     ds_all = nc.open_data()
    #     multiple = True
    # vars = vv
    # del vv
    # for vv in vars:
    with warnings.catch_warnings(record=True) as w:
        # print(vv)
        drop_variables = ["siglay", "siglev"]
        ds_xr = xr.open_dataset(ff, drop_variables=drop_variables, decode_times=False)
        long_name = ds_xr[vv].attrs["long_name"]
        ds1 = nc.from_xarray(ds_xr[vv])
        lon = ds1.to_xarray().lon.values
        lat = ds1.to_xarray().lat.values

        lon_min = float(lon.min())
        lon_max = float(lon.max())
        lat_min = float(lat.min())
        lat_max = float(lat.max())
        # handle longitudes over 180 appropriately
        if lon_min > 180:
            lon_min = lon_min - 360
        if lon_max > 180:
            lon_max = lon_max - 360
        extent = [lon_min, lon_max, lat_min, lat_max]
        session_info["extent"] = extent

        ds1.run()
        ds1.nco_command("ncks -d siglay,0,0")
        ds_xr = ds1.to_xarray()
        try:
            ds_xr = ds_xr.squeeze("siglay")
        except:
            pass
        ds1 = nc.from_xarray(ds_xr)
        ds1.subset(variable=vv)
        # ds1.tmean(align = "left")
        grid = pd.DataFrame({"lon": lon, "lat": lat})
        lon_max = grid["lon"].max()
        lon_min = grid["lon"].min()
        lat_max = grid["lat"].max()
        lat_min = grid["lat"].min()
        ds1.run()
        out_grid = nc.generate_grid.generate_grid(grid)
        nc.session.append_safe(out_grid)

        ds2 = ds1.copy()
        ds2.run()
        ds2.cdo_command(f"setgrid,{out_grid}")
        if lons is not None:
            ds2.to_latlon(lon=lons, lat=lats, res=res, method="nn")
        else:
            ds2.regrid(new_grid, method="nn")

        ds2.as_missing(0)
        ds2.run()
        df_mask = grid.assign(value=1)
        df_mask["lon"] = bin_value(df_mask["lon"], 0.25)
        df_mask["lat"] = bin_value(df_mask["lat"], 0.25)
        df_mask = df_mask.groupby(["lon", "lat"]).sum().reset_index()
        df_mask = df_mask.set_index(["lat", "lon"])
        ds_mask = nc.from_xarray(df_mask.to_xarray())
        os.system(f"cdo griddes {ds_mask[0]} > /tmp/mygrid")
        # open the text file text.txt and replace the string "generic" with "lonlat"
        with open("/tmp/mygrid", "r") as f:
            lines = f.readlines()

        # write line by line to /tmp/newgrid
        with open("/tmp/newgrid", "w") as f:
            for ll in lines:
                f.write(ll.replace("generic", "lonlat"))

        ds_mask.cdo_command(f"setgrid,/tmp/newgrid")
        # ds_mask.to_nc("/tmp/mask.nc")
        ds_mask.regrid(ds2, method="bil")
        # ds_mask.to_nc("/tmp/mask1.nc")
        ds_mask > 0
        # ds_mask.to_nc("/tmp/mask2.nc")
        os.remove("/tmp/mygrid")
        os.remove("/tmp/newgrid")
        ds_mask.as_missing(0)
        ds_mask.set_fill(-9999)
        ds2 * ds_mask
        ds2.set_longnames({ds2.variables[0]: long_name})
        ds2.subset(lon=[lon_min, lon_max], lat=[lat_min, lat_max])

        # if multiple is False:
        return ds2
    #     else:
    #         ds_all.append(ds2)
    # ds_all.merge("variables")
    # ds_all.run()
    # return ds_all


def fvcom_preprocess(
    variables=None, paths=None, lon_lim=None, lat_lim=None, res=0.05, out_dir=None
):
    """
    Preprocess FVCOM data for gridding and regridding.

    Parameters
    ----------
    variables : list
        List of variable names to process. This must be the names in the netCDF files.
    paths : list
        List of file paths to the FVCOM data files.
    lon_lim : list
        Minimum and maximum longitudes for regridding.
    lat_lim : list
        Minimum and maximum latitudes for regridding.
    res : list or float
        Resolution for regridding. This defaults to 0.05 degrees, which should be fine for point matchups.
    out_dir : str
        Output directory where processed data will be saved. If None, an error is raised.

    """

    # check if out_dir is None
    if out_dir is None:
        raise ValueError("out_dir must be specified")
    # check if out_dir exists
    if not os.path.exists(out_dir):
        os.makedirs(out_dir)
    
    # make sure variables is a list
    if not isinstance(variables, list):
        variables = [variables]

    with warnings.catch_warnings(record=True) as w:
        # print(vv)
        ds_all = nc.open_data()
        for vv in variables:
            print(f"Processing variable: {vv}")
            ds_vv = nc.open_data()
            for ff in tqdm(paths):
                ds = xr.open_dataset(
                    ff, drop_variables=["siglay", "siglev"], decode_times=False
                )
                ds_variables = ds.data_vars
                if vv not in ds_variables:
                    continue
                ds2 = fvcom_regrid(
                    ff=ff, new_grid=None, vv=vv, lons=lon_lim, lats=lat_lim, res=res
                )
                ds_vv.append(ds2)
            ds_vv.merge("time")
            ds_all.append(ds_vv)
        ff_out = out_dir + "/" + f"fvcom_values.nc"
        if os.path.exists(ff_out):
            os.remove(ff_out)
        ds_all.merge("variables")
        ds_all.to_nc(ff_out, zip=True)



# delete all objects starting with ds
for name in dir():
    if name.startswith("ds"):
        del globals()[name]
for name in dir():
    if name.endswith("mask"):
        del globals()[name]
nc.cleanup()
for ff in nc.session.get_safe():
    nc.session.remove_safe(ff)
nc.cleanup()


from oceanval.matchall import matchup
import pandas as pd
import glob
import os
import importlib

import nctoolkit as nc
import webbrowser

import re


def is_chunk(x):
    x = x.replace("\n", "")
    try:
        return re.findall(r"chunk_.[a-z]*", x)[0] == x
    except:
        return False


def add_chunks( dir=None):

    paths = glob.glob(f"{dir}/oceanval_report/notebooks/*.py")
    paths += glob.glob(f"oceanval_comparison/compare/notebooks/*.py")
    if dir is not None:
        paths += glob.glob(f"{dir}/*.py")

    for path in paths:
        # read file line by line
        # Using readlines()
        file1 = open(path, "r")
        Lines = file1.readlines()

        count = 0
        # Strips the newline character
        # generate new lines
        new_lines = []
        for line in Lines:
            if is_chunk(line):
                # get the file names
                chunk_file = line.replace("\n", "") + ".py"


                data_path = importlib.resources.files("oceanval").joinpath(f"data/{chunk_file}")

                # read the chunk file in line by line

                # Using readlines()
                file2 = open(data_path, "r")
                chunk_lines = file2.readlines()

                # add the chunk lines to the new lines
                for chunk_line in chunk_lines:

                    new_lines.append(chunk_line)

                # close the chunk file
                file2.close()

                count += 1
            else:
                new_lines.append(line)

        # close the original file
        file1.close()

        # write the new lines to the file
        file1 = open(path, "w")
        file1.writelines(new_lines)
        file1.close()



import nctoolkit as nc
import xarray as xr
import numpy as np
import subprocess
from oceanval.session import session_info


def bin_value(x, bin_res):
    return np.floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2


def extension_of_directory(starting_directory, exclude=[]):
    levels = session_info["levels_down"]

    new_directory = ""
    for i in range(levels):
        new_directory = new_directory + "/**"
    return new_directory + "/"


def is_latlon(ff):

    cdo_result = subprocess.run(
        f"cdo griddes {ff}",
        shell=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    return "lonlat" in cdo_result.stdout.decode("utf-8")


def get_extent(ff):
    # add docstring
    """ "
    Get the extent of a netcdf file

    Parameters
    ----------
    ff : str
        The path to the netcdf file

    Returns
    -------
    extent : list
        A list of the form [lon_min, lon_max, lat_min, lat_max]

    """

    ds = nc.open_data(ff)
    ds.subset(variables=ds.variables[0])
    ds.top()
    ds.tmean()
    ds.as_missing(0)
    ds_xr = ds.to_xarray()
    lon_name = [x for x in ds_xr.coords if "lon" in x][0]
    lat_name = [x for x in ds_xr.coords if "lat" in x][0]

    df = ds_xr.to_dataframe().reset_index()
    df = ds.to_dataframe().dropna().reset_index()
    df = df.rename(columns={lon_name: "lon", lat_name: "lat"})
    df = df.dropna()
    max_lon = df["lon"].max()
    if max_lon > 180:
        new_lon = df["lon"].values % 360
        new_lon = new_lon - 180
        lon_min = float(new_lon.min())
        lon_max = float(new_lon.max())
    else:
        lon_min = float(df.lon.min())
        lon_max = float(df.lon.max())
    lat_min = float(df.lat.min())
    lat_max = float(df.lat.max())
    return [lon_min, lon_max, lat_min, lat_max]

    lons = ds_xr[lon_name].values
    lons = lons.flatten()
    # as a unique, sorted list
    lons = list(set(lons))
    lats = ds_xr[lat_name].values
    lats = lats.flatten()
    # as a unique, sorted list
    lats = list(set(lats))
    lats.sort()
    lons.sort()
    lon_res = np.abs(lons[2] - lons[1])
    lat_res = np.abs(lats[2] - lats[1])
    print(lon_res)
    print(lat_res)
    ds = nc.open_data(ff)
    ds.subset(variables=ds.variables[0])
    ds.top()
    ds.tmax()
    ds.to_latlon(lon=[-180, 180], lat=[-90, 90], res=[lon_res, lat_res])
    # rename
    lon_min = df.lon.min()
    lon_max = df.lon.max()
    lat_min = df.lat.min()
    lat_max = df.lat.max()
    lons = [lon_min, lon_max]
    lats = [lat_min, lat_max]
    extent = [
        lons[0] - lon_res,
        lons[1] + lon_res,
        lats[0] - lat_res,
        lats[1] + lat_res,
    ]
    #
    return extent


def get_resolution(ff):
    ds = nc.open_data(ff, checks=False)
    var = ds.variables[0]
    ds = xr.open_dataset(ff)
    lon_name = [x for x in list(ds.coords) if "lon" in x][0]
    lat_name = [x for x in list(ds.coords) if "lat" in x][0]
    var_dims = ds[var].dims
    extent = get_extent(ff)
    if lon_name in var_dims:
        if lat_name in var_dims:
            n_lon = len(ds[lon_name])
            n_lat = len(ds[lat_name])
            lon_max = float(ds[lon_name].max())
            lon_min = float(ds[lon_name].min())
            lat_max = float(ds[lat_name].max())
            lat_min = float(ds[lat_name].min())
            lon_res = (lon_max - lon_min) / (n_lon - 1)
            lat_res = (lat_max - lat_min) / (n_lat - 1)
            return [lon_res, lat_res]
    else:
        # get the final two var_dims
        var_dims = var_dims[-2:]
        # lat should be the second
        n_lat = len(ds[var_dims[1]])
        n_lon = len(ds[var_dims[0]])
        lon_res = (extent[1] - extent[0]) / n_lon
        lat_res = (extent[3] - extent[2]) / n_lat
        return [lon_res, lat_res]




# %% [markdown] tags=["remove-cell"]
# ## Read in the data


# %% tags=["remove-input", "remove-cell"]
n_levels = definitions[variable].n_levels
if n_levels > 1: 
    layer_long = "sea surface"
else:
    layer_long = ""

ff = glob.glob(f"data_dir_value/oceanval_matchups/point/{layer}/{variable}/{point_source}/*_{variable}.csv")[0]

vv_source = os.path.basename(ff).split("_")[0]
vv_source_raw = vv_source
vv_source = vv_source.upper()
df = pd.read_csv(ff)
lon_max = lon_lim[1]
lon_min = lon_lim[0]
lat_max = lat_lim[1]
lat_min = lat_lim[0]
df = df.query(f"lon >= {lon_min} and lon <= {lon_max} and lat >= {lat_min} and lat <= {lat_max}").reset_index(drop = True) 
# drop duplicates
df = df.drop_duplicates().reset_index(drop = True)
ff_dict = f"data_dir_value/oceanval_matchups/point/{layer}/{variable}/matchup_dict.pkl"
point_time_res = ["year", "month", "day"]
point_time_res = [x for x in point_time_res if x in df.columns]

point_years = None
variable_formula = None
try:
    with open(ff_dict, "rb") as f:
        matchup_dict = pickle.load(f)
        min_year = matchup_dict["start"]
        max_year = matchup_dict["end"]
        point_time_res = matchup_dict["point_time_res"]
        if "point_years" in matchup_dict:
            point_years = matchup_dict["point_years"]
        if "model_variable" in matchup_dict:
            variable_formula = matchup_dict["model_variable"]
except:
    pass

if point_time_res is not None:
    if isinstance(point_time_res, str):
        point_time_res = [point_time_res]
    if "year" not in point_time_res:
        if "year" in df.columns:
            df = df.drop(columns = "year")
    if "month" not in point_time_res:
        if "month" in df.columns:
            df = df.drop(columns = "month")
    if "day" not in point_time_res:
        if "day" in df.columns:
            df = df.drop(columns = "day")
    df = df.drop_duplicates().reset_index(drop = True)
    grouping = [x for x in ["lon", "lat", "year", "depth", "day", "month"] if x in df.columns]
    df = df.groupby(grouping).mean().reset_index()

try:
    df = df.query("depth < 5").reset_index()
    grouping = [x for x in ["lon", "lat", "year", "day", "month"] if x in df.columns]
    df = df.groupby(grouping).mean().reset_index()
except:
    pass

df_locs = df.loc[:,["lon", "lat"]].drop_duplicates()
# bin to 0.01 resolution
df_raw = copy.deepcopy(df)




if len(point_time_res) > 0:
    if "year" in point_time_res:
        df = df.groupby(["lon", "lat", "year", "month"]).mean().reset_index()
    else:
        df = df.groupby(["lon", "lat",  "month"]).mean().reset_index()
        if "year" in df.columns:
            df = df.drop(columns = "year")
        if "day" in df.columns:
            df = df.drop(columns = "day")

grouping = ["lon", "lat", "year", "month"]
if bin_res is not None:
    lon_res = bin_res[0]
    lat_res = bin_res[1]
    # round to nearest bin_res using bin_value
    df = (
        df
        .assign(
            lon = lambda x: bin_value(x.lon, lon_res),
            lat = lambda x: bin_value(x.lat, lat_res)
        )
    )
    grouping = [x for x in grouping if x in df.columns]
    df = df.groupby(grouping).mean().reset_index()


available = len(df_raw) > 10


# %% tags=["remove-input", "remove-cell"]
# A function for generating the data source

def data_source(vv_source, vv_name):
    return vv_source


# %% tags=["remove-input"]
if available:
    intro = []
    
    if n_levels > 1:
        intro.append(f"This data was extracted from vertical profiles. Values from the **top 5m** were extracted from the database. This was compared with the model values from the sea surface level.")
    else:
        intro.append(f"This data was extracted from point measurements in the database.")
    
    model_variable = definitions[variable].model_variable    
    
    import pickle
    try:
        ff_dict = f"data_dir_value/oceanval_matchups/point/{layer}/{variable}/matchup_dict.pkl"
        with open(ff_dict, "rb") as f:
            matchup_dict = pickle.load(f)
            min_year = matchup_dict["start"]
            max_year = matchup_dict["end"]
            point_time_res = matchup_dict["point_time_res"]
    
        if min_year == max_year:
            intro.append(f"The model output was matched up with the observational data with model output from  the year **{min_year}**.")
        else:
            intro.append(f"The model output was matched up with the observational data with model output from the years **{min_year} to {max_year}**.")
        
        if point_time_res == ["year", "month", "day"]:
            intro.append(f"The model output was matched up precisely with the observational data for each day of the year in the years with data in both model and observations.")
        if point_time_res == ["month", "day"]:
            intro.append(f"The model output was matched up with the observational data for each day of the year. However, the year in the observational data was not considered, so the comparison is climatological.")
        if point_time_res == ["month"]:
            intro.append(f"The model output was matched up with the observational data for each month of the year. However, the year and day of month in the observational data was not considered, so the comparison is climatological.")
        if point_years is not None:
            point_start = point_years[0]
            point_end = point_years[1]
            if point_start > 1900:
                if point_start < point_end:
                    intro.append(f"The observational data was restricted to the years **{point_start} to {point_end}**.")
                else:
                    intro.append(f"The observational data was restricted to the year **{point_start}**.")
        
    except:
        if "year" in df_raw.columns:
            min_year = df_raw.year.min()
            max_year = df_raw.year.max()
            # coerce to int
            min_year = int(min_year)
            max_year = int(max_year)
        if min_year == max_year:
            intro.append(f"The model output was matched up with the observational data for the year **{min_year}**.")
        else:
            intro.append(f"The model output was matched up with the observational data for the years **{min_year} to {max_year}**.")
    
    
    
    md_basic(" ".join(intro).strip().replace("  ", " "))
    
    md(f"In total there were {len(df_raw)} values extracted from the observational database. The map below shows the locations of the matched up data for {vv_name}.", number = True)
    if variable_formula is not None:
        md_markdown(f"The following model output was used to compare with observational values: **{variable_formula}**.")
    else:
        md_markdown(f"The following model output was used to compare with observational values: **{model_variable}**.")
    if bin_res is not None:
        md_markdown(f"**Note**: the observational and model data were binned to a resolution of {bin_res[0]}° longitude by {bin_res[1]}° latitude and climatological monthly averages were calculated before analysis. This was carried out to reduce the influence of spatial bias on the validation statistics.") 
    md("**Note**: Individual vertical profiles with more than one observation in the top 5m were averaged to give a single observation for that profile.")

# %% tags=["remove-cell"]
# bottom 1% of observations
if available:
    bot_low = df.observation.quantile(0.001)
    df = df.query(f"observation >= {bot_low}")

# %% tags=["remove-input"]
# %%capture --no-display
# %%R -i df_locs -i variable -i available -i unit -i concise -w 500
if(available & concise == FALSE){
library(dplyr, warn.conflicts = FALSE)
library(ggplot2, warn.conflicts = FALSE)
library(stringr)
world_map <- map_data("world")
# get lon, lat limits from profile_mld
bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}

xlim = c(min(df_locs$lon), max(df_locs$lon))
ylim = c(min(df_locs$lat), max(df_locs$lat))


bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}

gg <- df_locs %>%
# final six months of the year
    ggplot()+
    geom_point(aes(lon, lat))+
    theme_gray(base_size = 14)+
    # add colour scale. Minimum zero, label 100, ">100"
    geom_polygon(data = world_map, aes(long, lat, group = group), fill = "grey60")+
    coord_fixed(xlim = xlim, ylim = ylim, ratio = 1.5, expand = FALSE) 

# figure out if lon minimum is less than -10
if( min(df_locs$lon) < -13 ){
    # add sensible labels for longitude and latitude

    gg <- gg +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 



}

    gg <- gg +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 

    # move legen

gg
}

# %% tags=["remove-input"]
if available and not concise:
    if n_levels > 1 and layer_select != "all": 
        md(f"**Figure {i_figure}:** Locations of matchups between simulated and observed {vv_name} in the top 5m of the water column.") 
    else:
        md(f"**Figure {i_figure}:** Locations of matchups between simulated and observed {vv_name}.") 

    i_figure = i_figure + 1

# %% tags=["remove-input"]
# %%capture --no-display
# %%R -i df -i variable -i unit -i layer_long -i vv_name -i available -w 800 
options(warn=-1)
options(warn=-1)
library(tidyverse)
library(stringr)
library(ggtext)
if(available){

bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}

if (("month" %in% colnames(df)) == FALSE){

df_map <- df %>%
    gather(variable, value, model:observation)
    df_map
# calculate the 98th percentil of the data
p98 = quantile(df_map$value, 0.98)
# cap the value at this
df_map$value = pmin(df_map$value, p98)

world_map <- map_data("world")

Layer <- str_to_title(layer_long)
name <- str_glue("{Layer} {vv_name} ({unit})")
# ensure everything is superscripted where necessary
# m-3 should be superscripted
# Note: this is for ggplot2
name <- str_replace_all(name, "m-([0-9]+)", "m<sup>-\\1</sup>")
name = str_replace(name, "/m\\^2", "m<sup>-2</sup>")
# fix /day
name = str_replace(name, "/day", "day<sup>-1</sup>")
# fix O_2
name <- str_replace(name, "O_2", "O<sub>2</sub>")
# CO2
name <- str_replace(name, "CO2", "CO<sub>2</sub>")


bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}


gg <- df_map %>%
    ggplot()+
    geom_point(aes(lon, lat, colour = value))+
    theme_gray(base_size = 14)+
    coord_fixed(ratio = 1.5, xlim = c(min(df$lon), max(df$lon)), ylim = c(min(df$lat), max(df$lat)), expand = FALSE)+
    labs(color = variable)+
    # log10
    theme(legend.position = "bottom", legend.title = element_markdown())  +
    facet_wrap(~variable)+
      scale_colour_viridis_c(
        # use unit for the label
        name = name,
                       guide = guide_colorbar(title.position = "bottom", title.hjust = 0.5, title.theme = element_markdown(angle = 0, size = 20, family = "Helvetica"))
  )+
    theme(

    legend.position = "bottom", legend.direction = "horizontal", legend.box = "horizontal", legend.key.width = unit(3.0, "cm"),
    legend.key.height = unit(1.0, "cm"))




gg <- gg + 
    geom_polygon(data = world_map, aes(long, lat, group = group), fill = "grey60")

# remove x and y axis names

# ditch the whitespace around the plot
gg <- gg + theme(plot.margin=unit(c(0,0,0,0),"cm"))
gg <- gg +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 
gg

}

}

# %% tags=["remove-input"]
if available:
    if "month" not in df.columns:
        md(f"**Figure {i_figure}:** Map of average {layer_long} {vv_name} in the model and observational datasets.")
        i_figure += 1

# %% tags=["remove-input"]
if available:
    if "month" in df_raw.columns:
        # summarize using md the number of observations in each month
        # get the minimum and maximum number in each month and report the month
        df_size = df_raw.groupby("month").size().reset_index()
        df_size.columns = ["month", "n"]
        n_min = df_size.n.min()
        n_max = df_size.n.max()
        month_min = list(df_size.query("n == @n_min").month.values)[0]
        months_max = list(df_size.query("n == @n_max").month.values)[0] 
        # convert to month names
        import calendar
        month_min = calendar.month_name[int(month_min)]
        months_max = calendar.month_name[int(months_max)] 

        # summarize using md

        fig_summary = [f"The number of observations in each month ranged from {n_min} in {month_min} to {n_max} in {months_max}."]

        if not concise:
            fig_summary.append(f"Figure {i_figure} below shows the distribution of observations in each month.")

        md(" ".join(fig_summary).strip().replace("  ", " "), number = True)

    if "month" in df_raw.columns:
        df_totals = (
            df_raw.groupby("month").size().reset_index()
            # rename
            .rename(columns = {0: "n"})
        )
    else:
        df_totals = pd.DataFrame({"month": ["All"], "n": [len(df_raw)]})



# %% tags=["remove-input"]
if available:
    bias_text = []

    bias_text.append(f"Figure {i_figure} below shows the bias between the model and observational data for {vv_name}.")
    bias_text.append(f"The bias is calculated as the model value minus the observational value, and it is shown for each month of the year.")

    md(" ".join(bias_text).strip().replace("  ", " "))

# %% tags=["remove-input"]
# %%capture --no-display
# %%R -i df -i variable -i unit -i layer_long -w 1000 -h 1200 -i vv_name -i available -i layer_select
if(available){
options(warn=-1)
# #%%R -i df -i variable -i unit -w 1600 -h 1000
options(warn=-1)
bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}

library(dplyr, warn.conflicts = FALSE)
library(ggplot2, warn.conflicts = FALSE)
library(stringr)
library(tidyverse)
world_map <- map_data("world")
# get lon, lat limits from profile_mld

xlim = c(min(df$lon), max(df$lon))
ylim = c(min(df$lat), max(df$lat))


bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}


df <- df %>%
    mutate(bias = model - observation) 

# calculate the absolate bias

df1 <- df %>%
    mutate(bias = abs(bias))
# calculate the 98th percentile of the absolute bias
bias_high <- df1$bias %>% quantile(0.98)
# cap the bias to +/1 98th percentile
df$bias[df$bias > bias_high] <- bias_high
df$bias[df$bias < -bias_high] <- -bias_high



plot_month <- FALSE
if("month" %in% colnames(df))
    plot_month <- TRUE

# # convert month number to month in profile_mld
if(plot_month){
    df <- df %>%
        arrange(month)
df$month <- factor(df$month, levels = df$month, labels = month.abb[df$month])
}
# df$month <- factor(df$month, labels = month.abb)

title <- str_glue("Bias in {layer_long} {vv_name} ({unit})")
# fix O_2
title <- str_replace(title, "O_2", "O<sub>2</sub>")
title <- str_replace(title, "CO2", "CO<sub>2</sub>")

out = str_glue("../../oceanval_results/{layer_select}/{variable}/{layer_select}_{variable}_bias.csv")

# # check directory exists for out
if (!dir.exists(dirname(out))){
    dir.create(dirname(out), recursive = TRUE)
}
df %>% write_csv(out)

# df.to_csv(out, index = False)

# export to csv

title = str_replace(title, "/m\\^3", "m<sup>-3")
title = str_replace(title, "/m\\^2", "m<sup>-2")
# fix  O_2
# replace pco2 with PCO_2
title = str_replace_all(title, "pco2", "pCO<sub>2</sub>")
title = str_replace(title, "pCO2", "pCO<sub>2</sub>")
title = str_replace(title, "O_2", "O<sub>2</sub>")
# 


title <- str_replace_all(title, "m-([0-9]+)", "m<sup>-\\1</sup>")


# not for ben
    plot_width <- 6.0
    if(df$month %>% n_distinct() < 9){
        plot_width <- 4.5
    }

gg <- df %>%
    ggplot()+
    geom_point(aes(lon, lat, colour = bias))+
    theme_dark(base_size = 24)+
    # add colour scale. Minimum zero, label 100, ">100"
    coord_fixed(xlim = xlim, ylim = ylim, ratio = 1.5, expand = FALSE) +
    # move legend to the top. Make it 3 cm wide
    # move legend title to the bottom and centre it
    scale_colour_gradient2(low = "blue", high = "red",
    limits = c(-bias_high, bias_high),
                       guide = guide_colorbar(title.position = "bottom", title.hjust = 0.5, title.theme = ggtext::element_markdown(angle = 0, size = 20, family = "Helvetica"))
  )+
    theme(
    legend.position = "bottom", legend.direction = "horizontal", legend.box = "horizontal", legend.key.width = unit(plot_width, "cm"),
    legend.key.height = unit(1.0, "cm"))+
    # set the legend title to bias
    labs(fill = title)


if (plot_month){
    #  option: figure out how many months are in the data
    # and wrap appropriately. this requires the w to be fixed in %%R. Not sure how to do this
    gg <- gg + facet_wrap(~month)
}

bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}
colour_lab <- str_glue("Model bias ({unit})")
colour_lab <- str_replace(colour_lab, "/m\\^3", "m<sup>-3</sup>")
colour_lab <- str_replace(colour_lab, "/m\\^2", "m<sup>-2</sup>")
colour_lab <- str_replace_all(colour_lab, "m-([0-9]+)", "m<sup>-\\1</sup>")
# fix /day
colour_lab <- str_replace(colour_lab, "/day", "day<sup>-1</sup>")
# fix O_2
colour_lab <- str_replace(colour_lab, "O_2", "O<sub>2</sub>")
colour_lab <- str_replace(colour_lab, "CO2", "CO<sub>2</sub>")

#
gg <- gg + labs(colour = title)

gg <- gg + 
    geom_polygon(data = world_map, aes(long, lat, group = group), fill = "grey70")

gg <- gg +
    # remove the x and y axis totally
    theme(axis.text.x = element_blank(), axis.text.y = element_blank(),
          axis.ticks.x = element_blank(), axis.ticks.y = element_blank(),
          axis.title.x = element_blank(), axis.title.y = element_blank()) +
    labs(x = "", y = "") 

    # move legen

gg
}

# %% tags=["remove-input"]
if available:
    if n_levels > 1 and layer_select != "all":
        md(f"**Figure {i_figure}**: Bias in {layer_long} {vv_name}. The bias is calculated as model - observation. The colour scale is from blue (negative bias) to red (positive bias). The colour scale is capped at the 98th percentile of the absolute bias. This is to avoid a few extreme outliers from dominating the colour scale. **Note:** values have been binned and averaged to the resolution of the model.") 
    else:
        md(f"**Figure {i_figure}**: Bias in {layer_long} {vv_name}. The bias is calculated as model - observation. The colour scale is from blue (negative bias) to red (positive bias). The colour scale is capped at the 98th percentile of the absolute bias. This is to avoid a few extreme outliers from dominating the colour scale.") 
    i_figure += 1

# create directory if non-existent, recursive
    if os.path.isdir("adhoc/tmp") == False:
        os.makedirs("adhoc/tmp")
    df_raw.to_csv("adhoc/tmp/df_raw.csv")
    df.to_csv("adhoc/tmp/df.csv")


# %% tags=["remove-input"]
if available:
    scatter_text = []
    scatter_text.append(f"Figure {i_figure} shows the distribution of {layer_long} {vv_name} observations in the model and observational datasets.") 

    md(" ".join(scatter_text).strip().replace("  ", " "))


# %% tags=["remove-input"]
# %%capture --no-display
# %%R -i df -i concise -i vv_name -i unit -w 1000 -h 1200 -i available
if(concise == FALSE & available){

bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}

library(tidyverse, warn.conflicts = FALSE)


x_lab <- str_glue("Model {vv_name} ({unit})")
y_lab <- str_glue("Observed {vv_name} ({unit})")


x_lab <- str_replace(x_lab, "/m\\^3", "m<sup>-3</sup>")
y_lab <- str_replace(y_lab, "/m\\^3", "m<sup>-3</sup>")
x_lab <- str_replace(x_lab, "/m\\^2", "m<sup>-2</sup>")
y_lab <- str_replace(y_lab, "/m\\^2", "m<sup>-2</sup>")
x_lab <- str_replace_all(x_lab, "m-([0-9]+)", "m<sup>-\\1</sup>")
y_lab <- str_replace_all(y_lab, "m-([0-9]+)", "m<sup>-\\1</sup>")
# fix O_2
x_lab <- str_replace(x_lab, "O_2", "O<sub>2</sub>")
y_lab <- str_replace(y_lab, "O_2", "O<sub>2</sub>")
x_lab <- str_replace(x_lab, "CO2", "CO<sub>2</sub>")
y_lab <- str_replace(y_lab, "CO2", "CO<sub>2</sub>")


if ("month" %in% colnames(df)){
df <- df %>%
# convert month number to name, e.g. 1=Jan
# do not use a factor
    mutate(month = month.abb[month]) %>%
    ungroup()

df <- df %>%
    mutate(month = "All months") %>%
    ungroup() %>%
    bind_rows(df)

df$month <- factor(df$month, levels = c("All months", month.abb))

}
# replace pco2 with pCO2 with superscript in x_lab
x_lab <- str_replace_all(x_lab, "co2", "CO<sub>2</sub>")
y_lab <- str_replace_all(y_lab, "co2", "CO<sub>2</sub>")
x_lab <- str_replace_all(x_lab, "CO2", "CO<sub>2</sub>")
y_lab <- str_replace_all(y_lab, "CO2", "CO<sub>2</sub>")
# fix /day
x_lab <- str_replace(x_lab, "/day", "day<sup>-1</sup>")
y_lab <- str_replace(y_lab, "/day", "day<sup>-1</sup>")
# fix O_2
x_lab <- str_replace(x_lab, "O_2", "O<sub>2</sub>")
y_lab <- str_replace(y_lab, "O_2", "O<sub>2</sub>")



gg <- df %>%
# final six months of the year
    ggplot()+
    geom_point(aes(model, observation))+
    theme_gray(base_size = 24)+
    labs(fill = title)+
    geom_abline()+
    geom_smooth(aes(model, observation), method = "lm", se = FALSE)+
    labs(x = x_lab, y = y_lab)+
    theme(axis.title.x = ggtext::element_markdown())+
    theme(axis.title.y = ggtext::element_markdown())

    # move legen

if ("month" %in% colnames(df)){
    gg <- gg + 
    facet_wrap(~month)
}
gg
}

# %% tags=["remove-input"]
if available:
    if concise is False:
        if n_levels > 1: 
            md(f"**Figure {i_figure}**: Simulated versus observed {vv_name} in the top 5m of the water column. The blue curve is a linear regression fit to the data, and the black line represents 1-1 relationship between the simulation and observations. The data has been averaged per model grid cell.") 
        else:
            if "year" not in df.columns:
                md(f"**Figure {i_figure}**: Simulated versus observed {vv_name}. The blue curve is a linear regression fit to the data, and the black line represents 1-1 relationship between the simulation and observations. The data has been averaged per model grid cell and month.")
        i_figure = i_figure + 1

# %% tags=["remove-input"]
# %%capture --no-display
# %%R -i vv_name -i unit -i concise -w 500  -i available

if(concise & available){
library(dplyr, warn.conflicts = FALSE)
library(ggplot2, warn.conflicts = FALSE)
library(stringr)

bin_value <- function(x, bin_res) {
	floor((x + bin_res / 2) / bin_res + 0.5) * bin_res - bin_res / 2
}

df <- read_csv("adhoc/tmp/df.csv")



x_lab <- str_glue("Model {vv_name} ({unit})")
y_lab <- str_glue("Observed {vv_name} ({unit})")
x_lab <- str_replace(x_lab, "/m\\^3", "m<sup>-3</sup>")
y_lab <- str_replace(y_lab, "/m\\^3", "m<sup>-3</sup>")
x_lab <- str_replace(x_lab, "/m\\^2", "m<sup>-2</sup>")
y_lab <- str_replace(y_lab, "/m\\^2", "m<sup>-2</sup>")
x_lab <- str_replace_all(x_lab, "m-([0-9]+)", "m<sup>-\\1</sup>")
y_lab <- str_replace_all(y_lab, "m-([0-9]+)", "m<sup>-\\1</sup>")
# fix /day
x_lab <- str_replace(x_lab, "/day", "day<sup>-1</sup>")
y_lab <- str_replace(y_lab, "/day", "day<sup>-1</sup>")

gg <- df %>%
# final six months of the year
    ggplot()+
    geom_point(aes(model, observation))+
    theme_gray(base_size = 14)+
    labs(fill = title)+
    geom_abline()+
    geom_smooth(aes(model, observation), method = "lm", se = FALSE)+
    labs(x = x_lab, y = y_lab)+
    theme(axis.title.x = ggtext::element_markdown())+
    theme(axis.title.y = ggtext::element_markdown())
    # move legen

gg
}

# %% tags=["remove-input"]
if available:
    if concise:
        md(f"**Figure {i_figure}**: Model vs observed {vv_name} for {layer_long} values. The blue line is a linear regression model fit to the data.")
        i_figure += 1
    if n_levels > 1: 
        md(f"## Summary statistics for sea surface {vv_name}")
    else:
        md(f"## Summary statistics for {vv_name}") 

# %% tags=["remove-input"]
if available and not concise:
    md(f"The overall ability of the model to predict the observed {vv_name} was assessed by calculating the average bias, the root mean square deviation (RMSD) and the correlation coefficient (R). The bias was calculated as the model value minus the observed value. The RMSD was calculated as the square root of the mean squared deviation. The correlation coefficient was calculated as the Pearson correlation coefficient between the model and observed values.") 
    md(f"This was calculated for each month and for the entire dataset. The results are shown in the tables below.")

# %% tags=["remove-input"]
if available:
    if variable not in [ "benbio"]:
        df_bias = (
            df_raw
            .assign(bias = lambda x: x.model - x.observation)
            .groupby("month")
            .mean()
            .reset_index()
            .loc[:,["month", "bias"]]
            # convert month number to name
            .assign(month = lambda x: x.month.apply(lambda y: calendar.month_abbr[y]))
        )
        # add average bias to df_bias as a separate row
        annual_bias = df_raw.model.mean() - df_raw.observation.mean() 
        df_bias = pd.concat([df_bias, pd.DataFrame({"month": ["All"], "bias": [annual_bias]})])

        # move the final row to the top
        df_bias = pd.concat([df_bias.iloc[[-1]], df_bias.iloc[:-1]])
    else:
        # only want annual
        df_bias = pd.DataFrame({"month": ["All"], "bias": [df_raw.model.mean() - df_raw.observation.mean()]})
    if variable not in [ "benbio"]:
        # now create an rmse dataframe
        df_rmse = (
            df_raw
            .assign(month = lambda x: x.month.apply(lambda y: calendar.month_abbr[y]))
            .groupby("month")
            .apply(lambda x: np.sqrt((x.model - x.observation).pow(2).mean()))
            .reset_index()
            .rename(columns={0: "rmse"})
        )
        # add average rmse to df_rmse as a separate row
        annual_rmse = np.sqrt(((df_raw.model - df_raw.observation).pow(2)).mean())
        df_rmse = pd.concat([df_rmse, pd.DataFrame({"month": ["All"], "rmse": [annual_rmse]})])
        # move the final row to the top
        df_rmse = pd.concat([df_rmse.iloc[[-1]], df_rmse.iloc[:-1]])
    else:
        # only want annual
        df_rmse = pd.DataFrame({"month": ["All"], "rmse": [np.sqrt(((df_raw.model - df_raw.observation).pow(2)).mean())]})
    # rename the month column to Month
    # merge the two dataframes
    df_table = copy.deepcopy(df_bias).merge(df_rmse)
    df_table = df_table.round(2)
    # create df_corr
    if variable not in [ "benbio"]:
        df_corr = (
            df_raw
            .groupby("month")
            .apply(lambda x: x.model.corr(x.observation))
            .reset_index()
            .rename(columns={0: "correlation"})
            .assign(month = lambda x: x.month.apply(lambda y: calendar.month_abbr[y]))
        )
        # add average correlation to df_corr as a separate row
        # calculate annual correlation using all data
        annual_corr = df_raw.model.corr(df_raw.observation)
        df_corr = pd.concat([df_corr, pd.DataFrame({"month": ["All"], "correlation": [annual_corr]})])
        # df_corr = df_corr.append({"month": "All", "correlation": annual_corr}, ignore_index=True)

        # move the final row to the top
        df_corr = pd.concat([df_corr.iloc[[-1]], df_corr.iloc[:-1]])
    else:
        # only want annual
        df_corr = pd.DataFrame({"month": ["All"], "correlation": [df_raw.model.corr(df_raw.observation)]})
    df_table = df_table.merge(df_corr)
    df_table = df_table.round(2)
    df_table = df_table.rename(columns={"month": "Month", "bias": "Bias", "rmse": "RMSD", "correlation": "Correlation"})
    df_table = df_table[["Month", "Bias", "RMSD", "Correlation"]]
    # change Month to Period
    df_table = df_table.rename(columns={"Month": "Time period"})

    if variable not in [ "benbio"]:
        # add commas to bias and rmse
        df_number = df_raw.groupby("month").count().reset_index().loc[:,["month", "observation"]]
    # convert month number to name
        df_number["month"] = df_number["month"].apply(lambda x: calendar.month_abbr[x])
        df_number = df_number.rename(columns={"month": "Time period", "observation": "Number of observations"})
    else:
        df_number = pd.DataFrame({"Time period": ["All"], "Number of observations": [len(df_raw)]})

    # add total number of observations
    annual_number = len(df_raw)
    if variable not in [ "benbio"]:
        df_number = pd.concat([df_number, pd.DataFrame({"Time period": ["All"], "Number of observations": [annual_number]})])
    # df_number = df_number.append({"Time period": "All", "Number of observations": annual_number}, ignore_index=True)
    df_table = df_table.merge(df_number)

    # include commas in the number of observations
    df_table["Number of observations"] = df_table["Number of observations"].apply(lambda x: "{:,}".format(x))

    # Now, we need to add the number of observations in df to df_table
    # a dataframe with them

    df_agg_numbers = (
        df
        .groupby("month")
        .size()
        .reset_index(name="n_observations")
    )
    # add total number of observations
    if bin_res is not None:
        total_n = len(df)
        df_agg_numbers = pd.concat([df_agg_numbers, pd.DataFrame({"month": ["All"], "n_observations": [total_n]})])
        # df_agg_numbers
        df_agg_numbers = df_agg_numbers.rename(columns={"month": "Time period", "n_observations": "Number of observations"})
        # convert month number to name
        df_agg_numbers["Time period"] = df_agg_numbers["Time period"].apply(lambda x: calendar.month_abbr[x] if x != "All" else "All")
        # include
        df_agg_numbers["Agg_Number"] = df_agg_numbers["Number of observations"].apply(lambda x: "{:,}".format(x))   
        # drop Number of observations
        df_agg_numbers = df_agg_numbers.drop(columns=["Number of observations"])
        df_agg_numbers = df_agg_numbers.merge(df_table, on = "Time period")
        # Number of observations is Agg_number with  (Number of observations))
        df_agg_numbers = df_agg_numbers.assign(**{"Number of observations": lambda x: x.Agg_Number + " (" + x["Number of observations"] + ")"})
        df_agg_numbers = df_agg_numbers.drop(columns=["Agg_Number"])
        # put All first
        df_agg_numbers = pd.concat([df_agg_numbers[df_agg_numbers["Time period"] == "All"], df_agg_numbers[df_agg_numbers["Time period"] != "All"]])

        df_display(df_agg_numbers)
    else:
        df_display(df_table)



# %% tags=["remove-input"]
if bin_res is not None:
    final_text = " Numbers in brackets show the number of raw unbinned observations."
else:
    final_text = ""

md(f"**Table {i_table}:** Average bias ({unit}) and root-mean square deviation ({unit}) for the model's {layer_long} {vv_name} for each month. The bias is calculated as model - observation. The average bias is calculated as the mean of the monthly biases. {final_text}")
i_table += 1


# %% tags=["remove-input"]
if available:
    md(f"A linear regression analysis of modelled and observed {vv_name} was performed. The modelled {vv_name} was used as the independent variable and the observed {vv_name} was used as the dependent variable. The results are shown in the table below.")

    md("The regression was carried out using the Python package statsmodels.")

# %% tags=["remove-input"]

# do a linear regression of model vs observed in df
if available:
    X = df.model.values
    Y = df.observation.values
    # linear regression using statsmodels
    import statsmodels.api as sm
    X = sm.add_constant(X)
    # make X and Y random numbers between 0 and 1
    X = sm.add_constant(X)
    model = sm.OLS(Y, X).fit()
    # get the slope and intercept
    intercept, slope = model.params
    # calculate the r squared
    r2 = model.rsquared
    # calculate the p value of the slope
    p = model.f_pvalue

    p = model.f_pvalue
    # put that in a dataframe
    df_stats = pd.DataFrame({"Slope": slope, "Intercept": intercept, "R2": r2, "P": p}, index = ["All"]).assign(Period = "All")
    # do this month by month append to df_stats

    for month in range(1, 13):
        try:
            X = df.query("month == @month").model.values
            Y = df.query("month == @month").observation.values
            X = sm.add_constant(X)
            model = sm.OLS(Y, X).fit()
            intercept, slope = model.params
            r2 = model.rsquared
            p = model.f_pvalue
            df_stats = pd.concat([df_stats, pd.DataFrame({"Slope": slope, "Intercept": intercept, "R2": r2, "P": p}, index = [month]).assign(Period = month)])
            df_stats.loc[df_stats.index[-1], "Period"] = calendar.month_abbr[month]
        except:
            pass
    # sort period appropriately, so All is first then ordered by month
    df_stats["Period"] = pd.Categorical(df_stats["Period"], [calendar.month_abbr[x] for x in range(1, 13)] + ["All"])
    # round p-value to 3 dp
    df_stats["P"] = df_stats["P"].round(5)
    # change P to p-value
    df_stats = df_stats.rename(columns={"P": "p-value"})
    # put Period first
    df_stats = df_stats[["Period", "Slope", "Intercept", "R2", "p-value"]]
    # 
    df_display(df_stats)

# %% tags=["remove-input"]
if available:
    md(f"**Table {i_table}:** Linear regression analysis of modelled and observed {vv_name}. The modelled {vv_name} was used as the independent variable and the observed {vv_name} was used as the dependent variable. The slope and intercept of the regression line are shown, along with the R<sup>2</sup> value and the p-value of the slope. Note: only months with sufficient values for a regression are shown.")
    i_table += 1 



import glob
import copy
import os
import warnings
import pickle
import pandas as pd
import time
import numpy as np
import nctoolkit as nc
import xarray as xr

from oceanval.fixers import tidy_warnings
from oceanval.utils import extension_of_directory, get_extent, is_latlon, get_resolution
from oceanval.session import session_info
from oceanval.parsers import Validator, definitions


def gridded_matchup(
    df_mapping=None,
    folder=None,
    var_choice=None,
    exclude=None,
    sim_start=None,
    sim_end=None,
    lon_lim=None,
    lat_lim=None,
    times_dict=None
):
    """
    Function to create gridded matchups for a given set of variables

    Parameters
    ----------
    df_mapping : pandas.DataFrame
        DataFrame containing the mapping between model variables and gridded observations
    folder : str
        Path to folder containing model data
    var_choice : list
        List of variables to create matchups for
    exclude : list
        List of strings to exclude from the file search
    sim_start : int
        Start year for model simulations
    sim_end : int
        End year for model simulations
    lon_lim : list
        Longitude limits for subsetting
    lat_lim : list
        Latitude limits for subsetting
    times_dict : dict
        Dictionary with file paths as keys and corresponding time DataFrames as values
    Returns
    -------
    None

    """

    start_cores = session_info["cores"]


    all_df = df_mapping
    # if model_variable is None remove from all_df
    good_model_vars = [x for x in all_df.model_variable if x is not None]

    all_df = all_df.query("model_variable in @good_model_vars").reset_index(drop=True)

    all_df = all_df.dropna()

    vars = []
    for x in definitions.keys:
        if definitions[x].gridded:
            vars.append(x)

    vars = [x for x in vars if x in var_choice]
    vars.sort()

    if len(vars) > 0:
        # first up, do the top

        mapping = dict()

        for vv in vars:
            vertical_gridded = definitions[vv].vertical_gridded
            climatology = definitions[vv].climatology
                # if global is not selected, stop
            # a dictionary for summarizing things
            var_dict = {}
            out_dir = session_info["out_dir"]
            out = glob.glob(
                out_dir + f"oceanval_matchups/gridded/{vv}/*_{vv}_surface.nc"
            )
            if len(out) > 0:
                if session_info["overwrite"] is False:
                    continue
            out_vertical = glob.glob(
                out_dir + f"oceanval_matchups/gridded/{vv}/*_{vv}_vertical.nc"
            )
            if len(out_vertical) > 0:
                if session_info["overwrite"] is False:
                    continue
            # figure out the data source
            #
            # check if this directory is empty
            dir_var = definitions[vv].gridded_dir 

            vv_source = definitions[vv].gridded_source

            #
            vv_name = definitions[vv].long_name

            print(
                f"Matching up surface {vv_name} with {vv_source.upper()} gridded data"
            )
            print("**********************")
            df = df_mapping.query("variable == @vv").reset_index(drop=True)

            if len(df) == 0:
                print("There appear to be no simulation files that meet your criteria.")
                continue

            mapping[vv] = list(df.query("variable == @vv").model_variable)[0]

            selection = []
            try:
                selection += mapping[vv].split("+")
            except:
                selection = selection

            patterns = set(df.pattern)
            if len(patterns) > 1:
                raise ValueError(
                    "Something strange going on in the string patterns. Unable to handle this. Bug fix time!"
                )
            pattern = list(patterns)[0]

            final_extension = extension_of_directory(folder)
            paths = glob.glob(folder + final_extension + pattern)

            for exc in exclude:
                paths = [
                    x for x in paths if f"{exc}" not in os.path.basename(x)
                ]


            all_years = []
            for ff in paths:
                all_years += list(times_dict[ff].year)

            all_years = list(set(all_years))
            n_years = len(all_years)

            sim_years = range(sim_start, sim_end + 1)
            sim_years = [x for x in all_years if x in sim_years]
            min_year = min(all_years)
            max_year = max(all_years)

            start = definitions[vv].gridded_start
            end = definitions[vv].gridded_end
            if start is not None:
                sim_years = [x for x in sim_years if x >= start]
            if end is not None:
                sim_years = [x for x in sim_years if x <= end]

            if len(sim_years) == 0:
                # specific error for glodap
                session_info["end_messages"] += [f"No simulation years found for {vv}. Please check start and end args!"]
                warnings.warn(f"No simulation years found for {vv}. Please check start and end args!")
                return None
            # now simplify paths, so that only the relevant years are used
            new_paths = []
            year_options = list(
                set(
                    pd.concat([x for x in times_dict.values()])
                    .loc[:, ["year", "month"]]
                    .drop_duplicates()
                    .groupby("year")
                    .size()
                    # must be at least 12
                    .pipe(lambda x: x[x >= 12])
                    .reset_index()
                    .year
                )
            )
            old_years = sim_years
            n_years = len(sim_years)
            sim_years = [x for x in sim_years if x in year_options]
            if n_years > 0:
                if len(sim_years) == 0:
                    raise ValueError(
                        f"No years in common between model and observation for {vv}. Please check start and end args!"
                    )
            month_sel = range(1, 13)
            if len(sim_years) == 0:
                sim_years = old_years
                month_sel = list(
                    set(
                        pd.concat([x for x in times_dict.values()])
                        .loc[:, ["year", "month"]]
                        .drop_duplicates()
                        .month
                    )
                )

            for ff in paths:
                if len([x for x in times_dict[ff].year if x in sim_years]) > 0:
                    new_paths.append(ff)

            paths = list(set(new_paths))
            paths.sort()

            var_dict["clim_years"] = [min(sim_years), max(sim_years)]

            # get the number of paths

            with warnings.catch_warnings(record=True) as w:

                new_paths = copy.deepcopy(paths)

                ds_model = nc.open_data(paths, checks=False)
                ds_model.subset(years=sim_years)
                ds_model.subset(variables=selection)
                if vertical_gridded is False:
                    ds_zz = nc.open_data(ds_model[0], checks = False)
                    ds_zz.subset(variables=selection)
                    ds_zz.run()
                    if ds_zz.contents.nlevels.values[0] > 1:
                        ds_model.cdo_command("topvalue")
                if session_info["as_missing"] is not None:
                    ds_model.as_missing(session_info["as_missing"])
                ds_model.merge("time")
                ds_model.tmean(
                    ["year", "month"], align="left"
                )

                # the code below needs to be simplifed
                # essentially anything with a + in the mapping should be split out
                # and then the command should be run for each variable

                var_unit = None
                ignore_later = []
                for vv in list(df.variable):
                    if "+" in mapping[vv]:
                        command = f"-aexpr,{vv}=" + mapping[vv]
                        ds_model.cdo_command(command)
                        drop_these = mapping[vv].split("+")
                        ds_contents = ds_model.contents
                        ds_contents = ds_contents.query(
                            "variable in @drop_these"
                        )
                        var_unit = ds_contents.unit[0]
                        ds_model.drop(variables=drop_these)
                        ignore_later.append(vv)

                        ds_model.run()
                        for key in mapping:
                            if key not in ignore_later:
                                if mapping[key] in ds_model.variables:
                                    ds_model.rename({mapping[key]: key})
                        if var_unit is not None:
                            ds_model.set_units(
                                {vv: var_unit}
                            )

                        ds_model.run()

            tidy_warnings(w)

            # figure out the start and end year
            with warnings.catch_warnings(record=True) as w:
                start_year = min(ds_model.years)
                end_year = max(ds_model.years)

                # Read in the monthly observational data
                thredds = definitions[vv].thredds
                if thredds == False:
                    if dir_var.endswith(".nc"):
                        vv_file = dir_var
                    else:
                        vv_file = nc.create_ensemble(dir_var)
                else:
                    vv_file = dir_var  # thredds URL
                recipe = definitions[vv].recipe
                
                # some special handling for occci files
                occci = False
                if recipe:
                    if vv_source.lower() == "occci":
                        new_files = []
                        for yy in sim_years:
                            for ff in vv_file:
                                if f"/{yy}/" in ff:
                                    new_files.append(ff)
                        vv_file = new_files
                        occci = True
                extracted = False
                if recipe:
                    if vv_source == "GLODAPv2.2016b":
                        print("Downloading GLODAPv2.2016b data")
                        ds_obs = nc.open_url(vv_file)
                        print("Download complete")
                        variable = definitions[vv].obs_variable
                        ds_obs.subset(variables=variable)
                        ds_obs.run()
                        # we need to set the time, because it does not exists
                        ds_obs.cdo_command(f"setreftime,1800-01-01,00:00:00,days -settaxis,2000-01-01,00:00:00,1month")
                        ds_obs.run()
                        thredds = False
                        extracted = True

                if not extracted:
                    if thredds:
                        ds_obs = nc.open_thredds(vv_file, checks=False)
                    else:
                        ds_obs = nc.open_data(
                            vv_file,
                            checks=False,
                        )
                bad_clim = False

                # use ncks to spatially subset to lon_lim and lat_lim
                if occci:
                    variable = definitions[vv].obs_variable
                    ds_obs.subset(variables=variable)
                    if lon_lim is not None:
                        ds_obs.crop(lon=lon_lim, lat=lat_lim)

                if vertical_gridded is False:
                    if thredds:
                        ds_zz = nc.open_thredds(ds_obs[0], checks = False)
                    else:
                        ds_zz = nc.open_data(vv_file, checks =False)
                    variable = definitions[vv].obs_variable
                    if variable == "auto":
                        variable = ds_zz.variables[0]
                    if ds_zz.contents.query("variable == @variable").nlevels.values[0] > 1:
                        if recipe:
                            ds_obs.top()
                        else:
                            ds_obs.cdo_command("topvalue")

                if definitions[vv].obs_variable != "auto":
                    ds_obs.subset(variables=definitions[vv].obs_variable)

                try:
                    min_obs_year = min(ds_obs.years)
                    max_obs_year = max(ds_obs.years)
                    if climatology is True:
                        if (min_obs_year < max_obs_year):
                            bad_clim = True
                    if climatology is False:
                        if (min_obs_year < max_obs_year):
                            if sim_start > max_obs_year or sim_end < min_obs_year:
                                session_info["end_messages"] += [f"No observation years found for gridded {vv}. Please check start and end args!"]
                                return None
                    if climatology is False:
                        ds_obs.subset(years=sim_years)
                    ds_obs.tmean(["year", "month"], align="left")
                    ds_obs.merge("time")
                    ds_obs.tmean(["year", "month"], align="left")
                except:
                    pass

                if climatology is False:
                    year_sel = [x for x in ds_model.years if x in ds_obs.years]
                    sim_years = year_sel
                    min_year = min(year_sel)
                    max_year = max(year_sel)
                    ds_obs.subset(years=year_sel)
                    ds_model.subset(years=year_sel)
                    ds_obs.run()

                obs_unit_multiplier = definitions[vv].obs_multiplier_gridded
                if obs_unit_multiplier != 1:
                    ds_obs *  obs_unit_multiplier
                obs_adder = definitions[vv].obs_adder_gridded
                if obs_adder != 0:
                    ds_obs + obs_adder
            
                if bad_clim:
                    raise ValueError(f"Observational data for {vv} appears to not be a climatology, but the climatology argument is set to True. Please fix this!") 

                if len(month_sel) < 12:
                    try:
                        months = ds.months
                        if len(months) > 1:
                            ds_obs.subset(months=month_sel)
                    except:
                        pass
                
                if definitions[vv].obs_variable != "auto":
                    ds_obs.subset(variables=definitions[vv].obs_variable)

                if climatology is False:
                    obs_years = ds_obs.years
                    min_obs_year = min(obs_years)
                    max_obs_year = max(obs_years)

            tidy_warnings(w)

            with warnings.catch_warnings(record=True) as w:
                if climatology is True:
                    ds_model.tmean("month", align="left")
                try:
                    if len(obs_years) == 1:
                        ds_obs.tmean("month", align="left")
                except:
                    pass

                if len(ds_obs) > 1:
                    ds_obs.merge("time")
                    ds_obs.run()

                lons = session_info["lon_lim"]
                lats = session_info["lat_lim"]
                if lons is None:
                    lons = [-180, 180]
                    lats = [-90, 90]

                # # figure out the lon/lat extent in the model

                lon_min_model = lons[0]
                lon_max_model = lons[1]
                lat_min_model = lats[0]
                lat_max_model = lats[1]

                regridding = "model_to_obs"

                ds_obs.rename({ds_obs.variables[0]: "observation"})
                ds_model.rename({ds_model.variables[0]: "model"})
                ds_model.run()
                ds_obs.run()

                # it is possible the years do not overlap, e.g. with satellite Chl
                if len(ds_model.times) > 12:
                    years1 = ds_model.years
                    years2 = ds_obs.years
                    all_years = [x for x in years1 if x in years2]
                    if len(all_years) != len(years1):
                        if len(all_years) != len(years2):
                            ds_obs.subset(years=all_years)
                            ds_model.subset(years=all_years)
                            ds_obs.run()
                            ds_model.run()
                if len(ds_obs) > 1:
                    ds_obs.merge("time")

                ds_obs.run()
                ds_model.run()

                if vertical_gridded is False:
                    contents = ds_model.contents
                    nlevels = contents.nlevels[0]
                    if nlevels > 1:
                        if recipe:
                            ds_obs.top()
                        else:
                            ds_obs.cdo_command("topvalue")
                try:
                    n_times = len(ds_obs.times)
                except:
                    n_times = 1
                try:
                    n_years = len(ds_obs.years)
                except:
                    n_years = 1

                ds_obs.run()
                ds_model.run()
                ds2 = ds_model.copy()
                if len(ds_model.times) == 12:
                    ds_model.set_year(2000)

                if len(ds_model.times) > 12:
                    # at this point, we need to identify the years that are common to both
                    ds_times = ds_model.times
                    try:
                        ds_years = [x.year for x in ds_times]
                        ds_months = [x.month for x in ds_times]
                    except:
                        ds_years = [int(str(x).split("T")[0].split("-")[0]) for x in ds_times ]
                        ds_months = [int(str(x).split("T")[0].split("-")[1]) for x in ds_times]

                    df_surface = pd.DataFrame(
                        {"year": ds_years, "month": ds_months}
                    )

                    ds_times = ds_obs.times
                    try:
                        ds_years = [x.year for x in ds_times]
                        ds_months = [x.month for x in ds_times]
                    except:
                        ds_years = [int(str(x).split("T")[0].split("-")[0]) for x in ds_times ]
                        ds_months = [int(str(x).split("T")[0].split("-")[1]) for x in ds_times]
                    df_obs = pd.DataFrame(
                        {"year": ds_years, "month": ds_months}
                    )
                    sel_years = list(
                        df_surface.merge(df_obs)
                        .groupby("year")
                        .count()
                        # only 12
                        .query("month == 12")
                        .reset_index()
                        .year.values
                    )
                    ds_model.subset(years=sel_years)
                    if n_years > 1: 
                        ds_obs.subset(years=sel_years)

                if len(ds_model.times) < 12:
                    sel_months = list(set(ds_model.months))
                    sel_months.sort()
                    if n_times > 1: 
                        ds_obs.subset(months=sel_months)

                # now do the surface
                ds_model_surface = ds_model.copy()
                contents = ds_model_surface.contents
                nlevels = contents.nlevels[0]
                if nlevels > 1:
                    ds_model_surface.cdo_command("topvalue")
                ds_obs_surface = ds_obs.copy()
                contents = ds_obs_surface.contents
                nlevels = contents.nlevels[0]
                if nlevels > 1: 
                    ds_obs_surface.cdo_command("topvalue")
                if regridding == "obs_to_model":
                    ds_obs_surface.regrid(ds_model_surface, method="bil")
                else:
                    ds_model_surface.regrid(ds_obs_surface, method="bil")

                n_obs_times = len(ds_obs_surface.times)
                if n_obs_times <= 1:
                    ds_model_surface.tmean()
                ds_obs_surface.append(ds_model_surface)

                if len(ds_model_surface.times) > 12:
                    ds_obs_surface.merge("variable", match=["year", "month"])
                else:
                    # run both
                    ds_obs_surface.run()
                    ds_model_surface.run()
                    ds_obs_surface.merge("variable", match="month")

                ds_obs_surface.set_fill(-9999)
                ds_mask = ds_obs_surface.copy()
                ds_mask.assign( mask_these=lambda x: -1e30 * ((isnan(x.observation) + isnan(x.model)) > 0), drop=True,)
                ds_mask.as_missing([-1e40, -1e20])
                ds_mask.run()
                ds_obs_surface + ds_mask

                out_file = (
                    session_info["out_dir"]
                    + f"oceanval_matchups/gridded/{vv}/{vv_source}_{vv}_surface.nc"
                )
                out_file_vertical = (
                    session_info["out_dir"]
                    + f"oceanval_matchups/gridded/{vv}/{vv_source}_{vv}_vertical.nc"
                )

                # check directory exists for out_file
                if not os.path.exists(os.path.dirname(out_file)):
                    os.makedirs(os.path.dirname(out_file))
                # remove the file if it exists
                if os.path.exists(out_file):
                    os.remove(out_file)
                ds_obs_surface.set_precision("F32")
                ds_model_surface = ds_obs_surface.copy()

                if lon_lim is not None and lat_lim is not None:
                    ds_model_surface.subset(lon=lon_lim, lat=lat_lim)

                ds_model_surface.run()
                if (
                    list(
                        ds_model_surface.contents.query(
                            "variable == 'model'"
                        ).long_name
                    )[0]
                    is None
                ):
                    ds_model_surface.set_longnames({"model": f"Model {vv_name}"})


                # unit may need some fiddling
                out1 = out_file.replace(
                    os.path.basename(out_file), "matchup_dict.pkl"
                )
                the_dict = {"start": min_year, "end": max_year}
                # write to pickle
                with open(out1, "wb") as f:
                    pickle.dump(the_dict, f)

                ds_model_surface.subset(lon=lons, lat=lats)
                ds_model_surface.run()

                lon_name = [x for x in ds_model_surface.to_xarray().coords if "lon" in x][0 ]
                lat_name = [x for x in ds_model_surface.to_xarray().coords if "lat" in x][0 ]
                ds_test = ds_model_surface.copy()
                ds_test.subset(variable = "observation")
                ds_test.tmean()
                ds_test.run()
                df_test = ds_test.to_dataframe().reset_index().dropna()
                lons = df_test[lon_name].values
                # handle lon > 180 properly
                lons = ((lons + 180) % 360) - 180
                lon_max = lons.max() 
                lon_min = lons.min()
                lat_max = df_test[lat_name].max()
                lat_min = df_test[lat_name].min()
                ds_model_surface.run()
#                ds_model_surface.subset(lon=[lon_min, lon_max], lat=[lat_min, lat_max])
                ds_model_surface.run()

                ds_model_surface.to_nc(out_file, zip=True, overwrite=True)
                out_file = out_file.replace(".nc", "_definitions.pkl")
                # save definitions
                with open(out_file, "wb") as f:
                    pickle.dump(definitions, f)


                if vertical_gridded:

                    if vertical_gridded is True:
                        levels = ds_obs.levels
                        if session_info["ds_depths"] != "z_level":
                            ds_model.vertical_interp(levels, thickness = session_info["ds_thickness"]) 
                        else:
                            ds_model.vertical_interp(levels , fixed = True)

                    if regridding == "obs_to_model":
                        ds_obs.regrid(ds_model, method="bil")
                    else:
                        ds_model.regrid(ds_obs, method="bil")
                    n_obs_times = len(ds_obs.times)
                    if n_obs_times <= 1:
                        ds_model.tmean()
                    ds_obs.append(ds_model)

                    if n_obs_times <= 1:
                            ds_obs.merge("variable")
                    else:
                        if len(ds_model.times) > 12:
                            ds_obs.merge("variable", match=["year", "month"])
                        else:
                            ds_obs.merge("variable", match="month")
                    ds_obs.set_fill(-9999)
                    ds_mask = ds_obs.copy()
                    ds_mask.assign( mask_these=lambda x: -1e30 * ((isnan(x.observation) + isnan(x.model)) > 0), drop=True,)
                    ds_mask.as_missing([-1e40, -1e20])
                    ds_mask.run()
                    ds_obs + ds_mask

                    out_file = (
                        session_info["out_dir"]
                        + f"oceanval_matchups/gridded/{vv}/{vv_source}_{vv}_surface.nc"
                    )

                    # check directory exists for out_file
                    if not os.path.exists(os.path.dirname(out_file_vertical)):
                        os.makedirs(os.path.dirname(out_file_vertical))
                    # remove the file if it exists
                    if os.path.exists(out_file_vertical):
                        os.remove(out_file_vertical)
                    ds_obs.set_precision("F32")
                    ds_model = ds_obs.copy()

                    if lon_lim is not None and lat_lim is not None:
                        ds_model.subset(lon=lon_lim, lat=lat_lim)

                    ds_model.run()
                    if (
                        list(
                            ds_model.contents.query(
                                "variable == 'model'"
                            ).long_name
                        )[0]
                        is None
                    ):
                        ds_model.set_longnames({"model": f"Model {vv_name}"})

                    regrid_later = False

                    # unit may need some fiddling
                    out1 = out_file.replace(
                        os.path.basename(out_file), "matchup_dict.pkl"
                    )
                    the_dict = {"start": min_year, "end": max_year}
                    # write to pickle
                    with open(out1, "wb") as f:
                        pickle.dump(the_dict, f)

                    ds_test = ds_model.copy()
                    ds_test.subset(variable = "observation")
                    ds_test.tmean()
                    ds_test.run()
                    df_test = ds_test.to_dataframe().reset_index().dropna()
                    lons = df_test[lon_name].values
                    lons = ((lons + 180) % 360) - 180
                    lon_max = lons.max()
                    lon_min = lons.min()
                    lat_max = df_test[lat_name].max()
                    lat_min = df_test[lat_name].min()
                    ds_model.subset(lon=[lon_min, lon_max], lat=[lat_min, lat_max])
                    ds_model.to_nc(out_file_vertical, zip=True, overwrite=True)
                out_file = out_file.replace(".nc", "_definitions.pkl")
                # save definitions
                with open(out_file, "wb") as f:
                    pickle.dump(definitions, f)

            tidy_warnings(w)

            out = (
                session_info["out_dir"]
                + f"oceanval_matchups/gridded/{vv}/{vv}_summary.pkl"
            )
            if not os.path.exists(os.path.dirname(out)):
                os.makedirs(os.path.dirname(out))
            var_dict["clim_years"] = [min(sim_years), max(sim_years)]
            with open(out, "wb") as f:
                pickle.dump(var_dict, f)
        
        session_info["cores"] = start_cores
        nc.options(cores = start_cores)

        return None
